/*******************************************************************************

  Author: Artur Szostak / Olav Torheim
  Email:  artur@alice.phy.uct.ac.za / torheim@ift.uib.no
  
  This sample code uses PSI and the IRQSIGNAL module to communicate with the
  HLT-RORC card.

********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <unistd.h>
#include <time.h>
#include "psi.h"
#include "psi_error.h"

#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include "irqsignal.h"
#include "fetchIRQline.h"

// Vendor and device IDs of the HLT-RORC card.
#define RORC_VENDOR_ID 0x10DC  // CERN-European Lab. for Particle Physics
#define RORC_DEVICE_ID 0x0033  // ALICE DDL to PCI interface (RORC)

// PCI address space mapped to main memory space:

typedef struct
{
	unsigned long CD800000;   // Input register
	unsigned long CD800004;   // Input register
	unsigned long CD800008;   // Input register
	unsigned long CD80000C;   // Input register
	unsigned long CD800010;   // Input register
	unsigned long CD800014;   // Input register
	unsigned long CD800018;   // Input register
	unsigned long CD80001C;   // Input register
	unsigned long CD800020;   // Input register
	unsigned long CD800024;   // Input register
	unsigned long CD800028;   // Input register
	unsigned long CD80002C;   // Input register
	unsigned long CD800030;   // Input register
	unsigned long CD800034;   // Input register
	unsigned long CD800038;   // Input register
	unsigned long CD80003C;   // Input register
	unsigned long CD800040;   // Input register
	unsigned long CD800044;   // Input register
	unsigned long CD800048;   // Input register
	unsigned long CD80004C;   // Input register
	unsigned long CD800050;   // Input register
	unsigned long CD800054;   // Input register
	unsigned long CD800058;   // Input register
	unsigned long CD80005C;   // Input register

} volatile Registers;


unsigned long dword;
	
unsigned long dma_base_address;
unsigned long dma_buffer_length;
unsigned long report_buffer;
unsigned long rd_ptr;
unsigned long irq_pulse_length;
unsigned long global_csr;
unsigned long event_length;
unsigned long pattern_generator;

tRegion rorc_region;
Registers* reg = NULL;

int counter = 1;


/* This is out signal handler which gets called every time the IRQ
   we register for gets fired.
 */
void signal_handler(int signal, siginfo_t* info, void* data)
{
	// Check that it was the HLT-RORC card that fired the IRQ.
	unsigned int irqstatus = reg->CD800054;
	if (irqstatus == 0x00000000)
	{
		printf("Someone else signaled an IRQ on the same line.\n");
		return;
	};

	if ((irqstatus == 0x00000001) || (irqstatus == 0x00000003))
	{
		// Increment rd_ptr
		if (rd_ptr==(dma_base_address + dma_buffer_length - 256)) // 0xCF000380
		{
			rd_ptr = dma_base_address;
			reg->CD800028 = 0xAFFED00F;
		}
		else
		{
			rd_ptr = rd_ptr + 256;
			reg->CD800028 = 0xAFFED00F;
		}

		reg->CD800030 = rd_ptr;

		printf("Got signal, counter = %d, rd_ptr = %8.8lX\n", counter++, rd_ptr);
	};

	if ((irqstatus == 0x00000002) || (irqstatus == 0x00000003))
	{
		printf("Got signal: end of event.\n");
	};
};

int my_signal; // Variable storing the signal number to get notification for
unsigned int HLT_RORC_irqline;  // Stores the IRQ line of the HLT-RORC card.


void cleanup(int signal) // Cleanup catches the SIGINT signal when the program is terminated with Ctrl-C
{
	irqsignal_status_t status;

	reg->CD800024 = 0x00000000; // Stop the pattern generator
	reg->CD800018 = 0x00000000; // Stop everything.
	reg->CD80001C = 0xFFFFFFFF; // Reset card.

	// Close mapping to BAR0 of HLT-RORC
	if ( CloseRORC(rorc_region, reg) == 1 )
	{
		printf("ERROR closing RORC");
	};


	// Unregister all signals that were registered with the irqsignal.o driver.
	status = IrqSignalUnregister(HLT_RORC_irqline, my_signal);
	if (status != IRQSIG_OK)
	{
		printf("ERROR: %s\n", IrqSignalStatusString(status) );
	};

	printf("Done.\n");
	exit(0);
};


/* Dumps memory at (*mem) to standard output in 32bit dwords.
   size number of dwords are dumped.
   width is the number of dwords to display per line.
 */
void DumpMem32(volatile unsigned long* mem, int size /* in dwords */, int width)
{
	int i;
	printf("size = %d, mem = 0x%8.8lX\n", size, (unsigned long)mem);
	if (mem == NULL) return;
	for (i = 0; i < size; i++)
	{
		if (i % width == 0)
		{
			printf("\n0x%8.8lX  ", (unsigned long)(mem + i));
		};
		printf("%8.8lX ", (unsigned long)mem[i]);
	};
	printf("\n");
};


/* Fills the memory at (*mem) in 32bit dwords chunks with the value of "value".
   size number of dwords are filled.
 */
void WriteMem32(volatile unsigned long* mem, int size /* in dwords */, unsigned long value)
{
	int i;
	int j;
	printf("size = %d, mem = 0x%8.8lX\n", size, (unsigned long)mem);
	if (mem == NULL) return;
	for (i = 0; i < size; i++)
	{
		mem[i] = value;

		// Wait a few clocks to prevent PCI burst write.
		// This is important since the simple logic does not implement burst write mode.
		for (j = 0; j < 10000; j++) ;   
	};
};


/* Prints the region status structure to the screen.
 */
void PrintRegionStatus(tRegionStatus stat)
{
	printf("Name: %s\n", (char*)&stat.name);
	printf("Users: %d\n", stat.users);
	printf("Address: 0x%8.8lX\n", stat.address);
	printf("Bus: %d\n", stat.bus);
	printf("Function: %d\n", stat.devfn);
	printf("Base: %d\n", stat.base);
	switch (stat.typ)
	{
	case _none_:    printf("Type: none\n"); break;
	case _mem_:     printf("Type: mem\n"); break;
	case _bigphys_: printf("Type: bigphys\n"); break;
	case _physmem_: printf("Type: physmem\n"); break;
	case _base_:    printf("Type: base\n"); break;
	case _baseIO_:  printf("Type: base IO\n"); break;
	case _config_:  printf("Type: config\n"); break;
	default:        printf("Type: INVALID VALUE!\n"); break;
	};
	printf("Size: %ld\n", stat.size);
};


/* Opens BAR0 of the RORC card and returns the memory mapped registers.
   NULL is returned if a error occured and a message is printed to screen.
 */
Registers* OpenRORC(tRegion* region)
{
	char name[1024];
	Registers* reg;
	PSI_Status status;
	tRegionStatus stat;

	// Open BAR0 of the HLT-RORC card.
	printf("\nOpening BAR0 on RORC card...\n");
	sprintf(name, "/dev/psi/vendor/0x%04hX/device/0x%04hX/0/base0",
		RORC_VENDOR_ID, RORC_DEVICE_ID);
	status = PSI_openRegion(region, name);
	if ( status != PSI_OK )
	{
		printf("ERROR: Unable to open BAR0 of RORC: ");
		printf("%s (%d)\n", PSI_strerror(status), (int)status);
		return NULL;
	};

	// Get the status information of the region.
	status = PSI_getRegionStatus(*region, &stat);
	if ( status != PSI_OK )
	{
		printf("ERROR: Unable to get status of BAR0 on RORC: ");
		printf("%s (%d)\n", PSI_strerror(status), (int)status);
		PSI_closeRegion(*region);
		return NULL;
	};
	printf("Region status:\n");
	PrintRegionStatus(stat);

	// Memory map the region into user space memory and get the pointer to
	// the mapped memory in the variable "reg"
	status = PSI_mapRegion(*region, (void**)&reg);
	if ( status != PSI_OK )
	{
		printf("ERROR: Unable to map BAR0 of RORC to virtual memory: ");
		printf("%s (%d)\n", PSI_strerror(status), (int)status);
		PSI_closeRegion(*region);
		return NULL;
	};
	printf("\n");

	return reg;
};

// Opens the configuration space of RORC, reading out from the Latency Timer Register and writing back a new value

void ConfigRORC(void)
{
	char name[1024];
	Registers* reg;
	PSI_Status status;
	tRegionStatus stat;
	tRegion* region;
	unsigned long i;

	// Open Configuration Space of the HLT-RORC card.
	printf("\nOpening configuration space on RORC card...\n");
	sprintf(name, "/dev/psi/vendor/0x%04hX/device/0x%04hX/0/config",
		RORC_VENDOR_ID, RORC_DEVICE_ID);
	status = PSI_openRegion(region, name);
	if ( status != PSI_OK )
	{
		printf("ERROR: Unable to open configuration space of RORC: ");
		printf("%s (%d)\n", PSI_strerror(status), (int)status);
	};

	// Get the status information of the region.
	status = PSI_getRegionStatus(*region, &stat);
	if ( status != PSI_OK )
	{
		printf("ERROR: Unable to get status of configuration space on RORC: ");
		printf("%s (%d)\n", PSI_strerror(status), (int)status);
		PSI_closeRegion(*region);
	};
	printf("Region status:\n");
	PrintRegionStatus(stat);

	// Reading out Latency Timer from Configuration Space

	status = PSI_read(*region, 0xC, _dword_,_dword_, &i);
	printf("Old Latency Timer value: %04hX", i);

	i = i & 0x00FF;
	i = i | 0x2000; // Setting latency timer to 64 = 0x40

	
	// Writing back new Latency Timer value
	
	printf(". New Latency Timer value: %04hX", i);
	status = PSI_write(*region, 0xC, _dword_,_dword_, &i);

	if ( status != PSI_OK )
	{
		printf("ERROR: Unable to read from configuration space: ");
		PSI_closeRegion(*region);
	};
	
	
	PSI_closeRegion(*region);
	printf("\n");

};

/* Closes the region allocated by OpenRORC.
   0 is returned on success and 1 of failure.
*/
int CloseRORC(tRegion region, Registers* reg)
{
	PSI_Status status;

	// Unmap the BAR0 region and close it:
	// PSI specifies that every mapped memory region mapped with PSI_mapRegion
	// must be unmapped with PSI_unmapRegion when one using it.
	printf("\nClosing RORC card...\n");
	status = PSI_unmapRegion(region, (void*)reg);
	if ( status != PSI_OK )
	{
		printf("ERROR: Unable to unmap BAR0 of RORC: ");
		printf("%s (%d)\n", PSI_strerror(status), (int)status);
		PSI_closeRegion(region);
		return 1;
	};

	status = PSI_closeRegion(region);
	if ( status != PSI_OK )
	{
		printf("ERROR: Unable to close BAR0 of RORC: ");
		printf("%s (%d)\n", PSI_strerror(status), (int)status);
		return 1;
	};
	printf("\n");

	return 0;
};


unsigned int GetIRQLine()
{
	PSI_Status status;
	unsigned char irqline;

	status = FetchIRQLineByVendor(RORC_VENDOR_ID, RORC_DEVICE_ID, 0x00, &irqline);
	if ( status != PSI_OK )
	{
		printf("ERROR: Unable fetch IRQ line number: ");
		printf("%s (%d)\n", PSI_strerror(status), (int)status);
		return -1;
	};

	return irqline;
};


int main (int argc, char **argv)
{
	int result;
	irqsignal_status_t status;
	struct sigaction action, old_action;
	
	HLT_RORC_irqline = GetIRQLine();
	printf("IRQ Line of HLT-RORC device = %d (0x%X)\n", HLT_RORC_irqline, HLT_RORC_irqline);

	// Initialise our signal handler.
	action.sa_sigaction = &signal_handler;
	sigemptyset(&action.sa_mask);
	action.sa_flags = SA_SIGINFO;
	action.sa_restorer = NULL;
	result = sigaction(SIGUSR1, &action, &old_action);
	if (result != 0)
	{
		perror("ERROR");
		return 1;
	};

	my_signal = SIGUSR1;  // Specify the signal number we want to receive for the above IRQ.

	// Register with the irqsignal.o driver.
	status = IrqSignalRegister(HLT_RORC_irqline, my_signal);
	if (status != IRQSIG_OK)
	{
		printf("ERROR: %s\n", IrqSignalStatusString(status) );
		return 1;
	};
	// Must end the program with ctrl-c, since we go into infinite loop
	struct sigaction close_program;
	close_program.sa_handler = cleanup;
	sigemptyset(&close_program.sa_mask);
	close_program.sa_flags = 0;
	result = sigaction(SIGINT, &close_program, 0); 
	if (result != 0)
	{
		perror("ERROR");
		return 1;
	};
	
	// Now we'd like to do some writing to the Configuration Space of HLT-RORC	
	ConfigRORC();


	// Then we want to access BAR0 of HLT-RORC		
	reg = OpenRORC(&rorc_region);
	if (reg == NULL) return 1;

	// Determining values for the local control registers of HLT-RORC
	dma_base_address = 0xFC800000;
	dma_buffer_length = 0x00000400;
	report_buffer = 0xFC800000;
	rd_ptr = 0xFC800100;
	irq_pulse_length = 0x00000008; // Kann koma heilt ned i 2
	global_csr = 0x00000091; // Enabling PG, DMA and IRQ transfers
	//global_csr = 0x00000011; // Enabling PG and DMA transfers
	event_length = 0x0000001B;
	pattern_generator = 0x00000000; // Skal vera 2

	//	Writing the determined values to the local control registers of HLT-RORC

	printf("\nWriting to CD800000: [0x%8.8lX] ...\n", dma_base_address);
	reg->CD800000 = dma_base_address;

	printf("\nWriting to CD800004: [0x%8.8lX] ...\n", dma_buffer_length);
	reg->CD800004 = dma_buffer_length;

	printf("\nWriting to CD800008: [0x%8.8lX] ...\n", report_buffer);
	reg->CD800008 = report_buffer;

	dword = 0x20000008; // rd_ptr_addr
	printf("\nWriting to CD80002C: [0x%8.8lX] ...\n", dword);
	reg->CD80002C = dword;

	printf("\nWriting to CD800030: [0x%8.8lX] ...\n", rd_ptr);
	reg->CD800030 = rd_ptr;

	printf("\nWriting to CD800058: [0x%8.8lX] ...\n", irq_pulse_length);
	reg->CD800058 = irq_pulse_length;

	printf("\nWriting to CD800018: [0x%8.8lX] ...\n", global_csr);
	reg->CD800018 = global_csr;

	printf("\nWriting to CD800020: [0x%8.8lX] ...\n", event_length);
	reg->CD800020 = event_length;
	
	dword = 0x00000001; // Resetting pattern generator
	printf("\nWriting to CD800024: [0x%8.8lX] ...\n", dword);
	reg->CD800024 = dword;

	printf("\nWriting to CD800024: [0x%8.8lX] ...\n", pattern_generator);
	reg->CD800024 = pattern_generator;
	
	printf("\nWriting to CD800028: [0x%8.8lX] ...\n", 0xAFFED00F);
	reg->CD800028 = 0xAFFED00F;
	
	printf("Listening for interrupts. Press Ctrl-C to abort.\n");

	// Go into an infinite loop and wait for signals generated by the irq signaler module.
	// Must stop the process with CTRL-C

	do
	{
		sleep(10);
	} while (1);

	return 0;
};



