/*********************************************************************************

   Author: Artur Szostak
   Email:  artur@alice.phy.uct.ac.za

   Declarations of for the irqsignal API library.

**********************************************************************************/

#ifndef IRQ_SIGNAL_API_H
#define IRQ_SIGNAL_API_H

#include "../src/irqsignal_mod.h"

#ifdef __cplusplus
extern "C"
{
#endif


/* The registration routine. Must be called if the current process wants to
   receive notification of a IRQ. The specified signal is sent to the calling
   process when the specified IRQ has occured, after calling this routine.

   Returns:
     IRQSIG_OK                 - Registered successfully.
     IRQSIG_IRQ_IN_USE         - The IRQ we are trying to register for is already in use
                                 by another driver whcih does not share its IRQ.
     IRQSIG_ALREADY_REGISTERED - The process has already registered this signal for the
                                 specified IRQ.
     IRQSIG_OUTOFMEM           - Could not allocate required system memory.
     IRQSIG_BADIRQ             - The IRQ number was invalid.
                                 i.e. irq number >= MAX_NUMBER_OF_IRQS
     IRQSIG_UNKNOWN_ERROR      - The kernel routine "request_irq" returned a unexpected result.
     IRQSIG_CANT_OPEN_DRIVER   - Could not open the driver. Perhaps the node /dev/irqsignal is missing.
 */
extern irqsignal_status_t IrqSignalRegister(int signal);

/* The unregistration routine. Must be called for every previous IrqSignalRegister,
   with the same irq and signal parameters. If the current process wants to do a batch
   unregister then it should use the routine IrqSignalUnregisterAll.

   Returns:
     IRQSIG_OK               - Unregistered successfully.
     IRQSIG_NOT_REGISTERED   - This process has not registered the specified signal
                               for this IRQ.
     IRQSIG_BADIRQ           - The IRQ number was invalid.
     IRQSIG_CANT_OPEN_DRIVER - Could not open the driver. Perhaps the node /dev/irqsignal is missing.
 */
extern irqsignal_status_t IrqSignalUnregister(int signal);

/* Unregisters all signal notification for the specified IRQ.

   Returns:
     IRQSIG_OK               - Unregistered successfully.
     IRQSIG_NOT_REGISTERED   - This process has not registered the specified signal
                               for this IRQ.
     IRQSIG_BADIRQ           - The IRQ number was invalid.
     IRQSIG_CANT_OPEN_DRIVER - Could not open the driver. Perhaps the node /dev/irqsignal is missing.
 */
extern irqsignal_status_t IrqSignalUnregisterAll(unsigned int irq);


#ifdef DEBUG

/* Removes all registration information for all processes for the specified
   IRQ only. This should only be used for debugging.

   Returns:
     IRQSIG_OK               - Successful reset.
     IRQSIG_BADIRQ           - The IRQ number was invalid.
     IRQSIG_CANT_OPEN_DRIVER - Could not open the driver. Perhaps the node /dev/irqsignal is missing.
 */
extern irqsignal_status_t IrqSignalReset(unsigned int irq);

/* Removes all registration information for all processes. This should only
   be used for debugging.

   Returns:
     IRQSIG_OK               - Successful reset.
     IRQSIG_CANT_OPEN_DRIVER - Could not open the driver. Perhaps the node /dev/irqsignal is missing.
 */
extern irqsignal_status_t IrqSignalResetAll(void);

/* Clears all internal structures and resets the module usage count to 1.

   Returns:
     IRQSIG_OK               - Successful reset.
     IRQSIG_CANT_OPEN_DRIVER - Could not open the driver. Perhaps the node /dev/irqsignal is missing.
 */
extern irqsignal_status_t IrqSignalHardReset(void);

/* Used for testing and debugging. This routine simulates a IRQ request in the
   irqsignal kernel module.

   Returns:
     IRQSIG_OK                 - Routine exited successfully. Does not indicate
                                 a succesfull test though. The signal dispatcher
                                 still needs to run properly.
     IRQSIG_BADIRQ             - The IRQ number was invalid.
     IRQSIG_CANT_SCHEDULE_TASK - Something went wrong when trying to schedule the
                                 signal dispatcher task.
     IRQSIG_CANT_OPEN_DRIVER   - Could not open the driver. Perhaps the node /dev/irqsignal is missing.
 */
extern irqsignal_status_t IrqSignalTest(unsigned int irq);

#endif // DEBUG


/* Returns a status message correpsonding to the given status code. If the code is unknown
   then we return the result of strerror(status).
 */
extern char* IrqSignalStatusString(irqsignal_status_t status);


#ifdef __cplusplus
};
#endif

#endif // IRQ_SIGNAL_API_H
