/*********************************************************************************

   Author: Artur Szostak
   Email:  artur@alice.phy.uct.ac.za

   Implementation of the irqsignal API library.

**********************************************************************************/

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>

#include "../../include/irqsignal.h"

#define DRIVER_NODE "/dev/irqsignal"


int drivernode = -1;  // file handle to the driver.
int registered_exit_routine = 0;  // Boolean value specifying if we registered the exit routine or not.


/* Close the driver if it has been already opened.
 */
void close_driver(void)
{
	if (drivernode != -1)
	{
		close(drivernode);
		drivernode = -1;
	};
};


/* Try to open the driver. If it is already open then we do nothing.
   IRQSIG_OK is returned if the open succeeds otherwise we return IRQSIG_CANT_OPEN_DRIVER
   and the details are returned by the system in "errno".
 */
irqsignal_status_t open_driver(void)
{
	if (drivernode == -1)
	{
		drivernode = open(DRIVER_NODE, O_RDWR);  // open in read/write mode.
		if (drivernode == -1)
			return IRQSIG_CANT_OPEN_DRIVER;
		// Register the exit routine to clean up properly.
		if (registered_exit_routine == 0)
		{
			if ( atexit(&close_driver) != 0 )
			{
				close(drivernode);
				drivernode = -1;
				return IRQSIG_CANT_OPEN_DRIVER;
			};
		};
	};
	return IRQSIG_OK;
};


///////////////////////////////////////////////////////////////////////////////
// Implementation of API routines. Described in the header: "irqsignal.h"

irqsignal_status_t IrqSignalRegister(unsigned int irq, int signal)
{
	irqsignal_status_t status;
	irqsignal_register_info_t info;
	info.irq = irq;
	info.signal = signal;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, IRQSIGNAL_REGISTER, &info);
};


irqsignal_status_t IrqSignalUnregister(unsigned int irq, int signal)
{
	irqsignal_status_t status;
	irqsignal_register_info_t info;
	info.irq = irq;
	info.signal = signal;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, IRQSIGNAL_UNREGISTER, &info);
};


irqsignal_status_t IrqSignalUnregisterAll(unsigned int irq)
{
	irqsignal_status_t status;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, IRQSIGNAL_UNREGISTER_ALL, irq);
};


#ifdef DEBUG

irqsignal_status_t IrqSignalReset(unsigned int irq)
{
	irqsignal_status_t status;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, IRQSIGNAL_UNREGISTER_ALL, irq);
};


irqsignal_status_t IrqSignalResetAll(void)
{
	irqsignal_status_t status;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, IRQSIGNAL_UNREGISTER_ALL);
};


irqsignal_status_t IrqSignalHardReset(void)
{
	irqsignal_status_t status;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, IRQSIGNAL_HARDRESET);
};


irqsignal_status_t IrqSignalTest(unsigned int irq)
{
	irqsignal_status_t status;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, IRQSIGNAL_TEST, irq);
};

#endif // DEBUG


char* IrqSignalStatusString(irqsignal_status_t status)
{
	switch (status)
	{
	case IRQSIG_OK:
		return "Successful";
	case IRQSIG_UNKNOWN_ERROR:
		return "Unknown error.";
	case IRQSIG_NO_SUCH_IOCTRL:
		return "Unrecognised ioctl number.";
	case IRQSIG_OUTOFMEM:
		return "System ran out of memory.";
	case IRQSIG_BADIRQ:
		return "Invalid IRQ number.";
	case IRQSIG_ALREADY_REGISTERED:
		return "Process already registered this signal for the specified IRQ.";
	case IRQSIG_IRQ_IN_USE:
		return "IRQ line in use by another driver.";
	case IRQSIG_NOT_REGISTERED:
		return "Process has not registered this signal for the specified IRQ.";
	case IRQSIG_CANT_SCHEDULE_TASK:
		return "Can not schedule the dispatcher task.";
	case IRQSIG_CANT_OPEN_DRIVER:
		return "Can not open driver file.";
	default:
		return strerror(status);
	};
};
