--
-- Authors: Artur Szostak / Olav Torheim
-- Email:  artur@alice.phy.uct.ac.za | artursz@iafrica.com / torheim@ift.uib.no
-- 
-- Asserting irq_generate for one clock cycle makes the IRQ signaler generate an interrupt.
-- The interrupt signal is active until the status register is cleared or the counter has reached pulse length 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;


entity IRQ_signaler is
  GENERIC (
    irq_status_in_ptr      : integer RANGE 0 TO 255 := 0;
    pulse_length_in_ptr     : integer RANGE 0 TO 255 := 4
  );
	port
	(
		-- PCI core signals
		clock   : in  std_logic; -- PCI clock.
		resetn   : in  std_logic; -- PCI core reset signal.
		
		-- IRQ signals
		irq_generate : in std_logic; -- Signalizes that an interrupt is to be generated
		irq_status_read : in std_logic; -- If asserted then a target read has succeeded
		irq_line : out std_logic; -- This must connect to the IRQ line of the PCI core.

		-- Reason bits for raising an IRQ.
		dma_buffer_full : in std_logic;
		end_of_event    : in std_logic;

		-- config IF
		conf_addr    : IN  std_logic_vector(7 DOWNTO 0);
  		conf_in      : IN  std_logic_vector(31 DOWNTO 0);
		conf_out     : OUT std_logic_vector(31 DOWNTO 0);
		conf_we      : IN  std_logic		
	);
end IRQ_signaler;


architecture IRQ_signaler_arch of IRQ_signaler is

	-- The states of the state machine used to signal IRQ's
	type SM_states is (SM_IDLE, SM_PULSE);
	signal state : SM_states;
	
	-- State machine used to clear interrupt flags after being read from host
	type CLEAR_states is (CLEAR_IDLE, CLEAR_START, CLEAR_STOP);
	signal cstate : CLEAR_states;

	SIGNAL we  : std_logic;
  
	signal counter : std_logic_vector (31 downto 0); -- counter register.
	
	-- The two registers for clock delay and pulse length.
	signal pulse_length : std_logic_vector (31 downto 0); -- pulse length register.
	signal irq_status : std_logic_vector (31 downto 0); -- irq status register.	

begin

    ---------------------------------------------------------------------------
    config_if:
    ---------------------------------------------------------------------------
    we    <= '1' WHEN (conf_we='1' AND conf_addr = pulse_length_in_ptr) ELSE '0'; -- write_pulse_length
    conf_out <= pulse_length       WHEN (conf_addr = pulse_length_in_ptr) ELSE
                irq_status WHEN (conf_addr = irq_status_in_ptr) ELSE                
                (OTHERS => '0');



	-- A process for each of the pulse_length and status registers.

	pulselength_register : process (clock, resetn, conf_in)
	begin
		if resetn = '0' then
			pulse_length <= x"00000000";
		elsif clock'event and clock = '1' then  -- wait for raising clock edge. 
			if we = '1' then -- write_control
				pulse_length <= conf_in;  -- Write to the register if write signal received.
			end if;
		end if;
	end process;

	irqstatus_register : process (clock, resetn, conf_in)
	begin
		if resetn = '0' then
			irq_status <= x"00000000";
		elsif clock'event and clock = '1' then  -- wait for raising clock edge. 
			if (irq_generate='1') then
				-- Flagging that an interrupt has been requested. Needed in a shared interrupt environment.
				if dma_buffer_full = '1' then
					irq_status(0) <= '1'; -- Least significant bit for host buffer full.
				end if;
				if end_of_event = '1' then
					irq_status(1) <= '1'; -- Second least significant bit for end of event.
				end if;
			elsif (cstate = CLEAR_STOP) then
				-- Clears the interrupt status bits after reading it
				irq_status <= (others => '0');
			end if;
		end if;
	end process;

	clear_state_machine : process (clock, resetn, conf_in)
	begin
		if resetn = '0' then
			cstate <= CLEAR_IDLE;
		elsif clock'event and clock = '1' then  -- wait for raising clock edge. 
			case cstate is
				when CLEAR_IDLE =>
					if ((conf_addr = irq_status_in_ptr) AND (irq_status(0)='1' or irq_status(1)='1')) then
						cstate <= CLEAR_START;
					else
						cstate <= CLEAR_IDLE;
					end if;
				when CLEAR_START =>
					if (conf_addr=irq_status_in_ptr AND irq_status_read='1') then
						cstate <= CLEAR_STOP;
					else
						cstate <= CLEAR_START;
					end if;
				when CLEAR_STOP =>
					cstate <= CLEAR_IDLE;
			end case;
		end if;
	end process;


	state_machine : process (clock, resetn, state, counter, irq_generate)
	begin
		if resetn = '0' then
			-- Initialise the state machine with default values.
			irq_line <= '1'; -- Keep IRQ line off by default. For some reason Altera uses '1' for off.
			counter <= x"00000000";
			state <= SM_IDLE;
		elsif clock'event and clock = '1' then  -- wait for raising clock edge. Do everything on the rising edge.

			case state is
				when SM_IDLE => -- Idle state: we are waiting for an enable signal.
					irq_line <= '1'; -- Keep IRQ line off while idle.
					if irq_generate = '1' then 
							-- Got the enable signal so we need to start the counter and begin the delay phase.
							counter <= x"00000001";
								if not (pulse_length = x"00000000") then -- Dont bother if the pulse length is also zero.
									irq_line <= '0';  -- Going to pulse state.
									state <= SM_PULSE;
								end if;
					end if;

				when SM_PULSE =>  -- In the pulse phase we keep the IRQ line asserted for as long as
												-- specified by the pulse_length. When the counter reaches the
												-- pulse_length value then we can move back to the IDLE state.
						if counter = pulse_length then
								irq_line <= '1';  -- changing state.
								state <= SM_IDLE;
						else
							if cstate = CLEAR_STOP then  -- If the ISR has responded by reading the status register
							                                                -- then we need to deassert the irq line 
							                                                -- and change to the idle state.
								irq_line <= '1';  -- changing state.
								state <= SM_IDLE;
							else
								irq_line <= '0';  -- Keep IRQ line asserted while in pulse state.
							end if;
							counter <= counter + x"00000001";	
						end if;
			end case;

		end if;
	end process;

end; 


