--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : csr_block.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de>
-- Company    : 
-- Platform   : 
-------------------------------------------------------------------------------
-- Description:
--
-- Global control and status registes for HLT-RORC
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author    Description
--
-- 07.07.2004  1.0      tilsner   created
--
-- 24.06.2005  1.1      O. Torheim  Added control bit for IRQ signaling
--=============================================================================

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;
USE ieee.std_logic_arith.ALL;

ENTITY csr_block IS
  GENERIC (
    global_csr : integer range 0 TO 255 := 0;
    global_rst : integer range 0 TO 255 := 4;
    fclear_cmd : integer RANGE 0 TO 255
  );
  PORT (
    clk         : IN std_logic;
    rstn        : IN std_logic;

    -- IF to logic
    s_rst       : OUT std_logic;
    busy_en     : OUT std_logic;
    pg_en       : OUT std_logic;
    ram_en      : OUT std_logic;
    ddl_en      : OUT std_logic;
    dma_en      : OUT std_logic;
    irq_en      : OUT std_logic;    
    fifo_clear  : OUT std_logic;
    dma_busy    : IN std_logic;
    fifo_usedw  : IN std_logic_vector(7 DOWNTO 0);
    fifo_full   : IN std_logic;
    fifo_empty  : IN std_logic;
    hostbuf_full: IN std_logic;
    pci_64bit   : IN std_logic;
    
    -- config IF
    conf_addr   : IN  std_logic_vector(7 DOWNTO 0);
    conf_in     : IN  std_logic_vector(31 DOWNTO 0);
    conf_out    : OUT std_logic_vector(31 DOWNTO 0);
    conf_we     : IN  std_logic
  );

END csr_block;



ARCHITECTURE behaviour OF csr_block IS

  TYPE fsm_state_type IS (idle, reset);
  SIGNAL fsm_state : fsm_state_type;
  
  SIGNAL we            : std_logic_vector(2 DOWNTO 0);
  SIGNAL global_ctrl   : std_logic_vector(7 DOWNTO 0);
  SIGNAL global_errors : std_logic_vector(3 DOWNTO 0);
  SIGNAL output_sig    : std_logic_vector(31 DOWNTO 0);
  SIGNAL rst_count     : std_logic_vector(3 DOWNTO 0);
  SIGNAL count_clr     : std_logic;
  SIGNAL count_en      : std_logic;
  
  BEGIN

    ---------------------------------------------------------------------------
    config_if:
    ---------------------------------------------------------------------------
    we(0)      <= '1' WHEN (conf_we='1' AND conf_addr = global_csr) ELSE '0';
    we(1)      <= '1' WHEN (conf_we='1' AND conf_addr = global_rst) ELSE '0';
    we(2)      <= '1' WHEN (conf_we='1' AND conf_addr = fclear_cmd) ELSE '0';
    conf_out   <= output_sig WHEN (conf_addr = global_csr) ELSE (OTHERS => '0');

    output_sig <= global_errors & "0000000" & pci_64bit & hostbuf_full &
                  dma_busy & fifo_empty & fifo_full & fifo_usedw & global_ctrl;

    
    -------------------------------------------------------------------------------
    global_cs_register:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        global_ctrl <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (fsm_state = reset) THEN
          global_ctrl <= (OTHERS => '0');
        ELSIF (we(0)='1') THEN
          global_ctrl <= conf_in(7 DOWNTO 0);
        ELSE
          global_ctrl <= global_ctrl;
        END IF;
      END IF;  
    END PROCESS;

    
    ---------------------------------------------------------------------------    
    global_error_register:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        global_errors <= (OTHERS => '0');
      ELSIF (clk'event AND clk = '1') THEN
        IF (we(2)='1') THEN
          global_errors <= (OTHERS => '0');
        ELSIF (fifo_full='1') THEN
          global_errors(0) <= '1';
        END IF;
      END IF;
    END PROCESS;

  
    ---------------------------------------------------------------------------
    global_syn_reset:
    ---------------------------------------------------------------------------      
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        fsm_state <= idle;
      ELSIF (clk'event and clk = '1') THEN
        CASE fsm_state IS
          WHEN idle =>
            IF (we(1)='1') THEN
              fsm_state <= reset;
            ELSE
              fsm_state <= idle;
            END IF;

          WHEN reset =>
            IF (rst_count = 5) THEN
              fsm_state <= idle;
            ELSE
              fsm_state <= reset;
            END IF;
          END CASE;
        END IF;
      END PROCESS;

    ---------------------------------------------------------------------------
    reset_counter:
    ---------------------------------------------------------------------------      
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        rst_count <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (count_clr = '1') THEN
          rst_count <= (OTHERS => '0');
        ELSIF (count_en = '1') THEN
          rst_count <= rst_count + '1';
        ELSE
          rst_count <= rst_count;
        END IF;
      END IF;
    END PROCESS;
    
    count_clr <= '1' WHEN (fsm_state = idle) ELSE '0';
    count_en  <= '1' WHEN (fsm_state = reset) ELSE '0';


    
    -- outputs
    s_rst      <= '1' WHEN (fsm_state = reset) ELSE '0';  -- synchronous reste
    fifo_clear <= we(2);


    pg_en   <= global_ctrl(4);          -- enable pattern generator
    ram_en  <= global_ctrl(5);          -- enable data transfer from RAM    
    ddl_en  <= global_ctrl(6);          -- enable data transfer from DDL

    dma_en  <= global_ctrl(0);          -- enable DMA transfers
    busy_en <= global_ctrl(1);          -- enable generation of busy signal
    
    irq_en <= global_ctrl(7);					-- OT 18.02.05: enable IRQ transfers
    


  END behaviour;
