--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : ddl_if.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de>
-- Company    :
-- Platform   :
-------------------------------------------------------------------------------
-- Description:
--
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author      Description
--
-- 07.07.2004  1.0      tilsner     created
--
-- 09.03.2005  1.1      A. Szostak  conf_addr is no longer compared to
--                                  hardcoded values.
--=============================================================================

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;

ENTITY ddl_if IS

  GENERIC (
    commd_reg_addr : integer RANGE 0 TO 255 := 0;
    ctstw_reg_addr : integer RANGE 0 TO 255 := 4;
    festw_reg_addr : integer RANGE 0 TO 255 := 8;
    dtstw_reg_addr : integer RANGE 0 TO 255 := 12;
    ifstw_reg_addr : integer RANGE 0 TO 255 := 16
  );
  
  PORT (
    clk       : IN std_logic;
    rstn      : IN std_logic;

    -- DDL interface
    roCLK     : OUT std_logic;
    roRST_N   : OUT std_logic;
    riLD_N    : IN  std_logic;
    riLF_N    : IN  std_logic;
    roBSY_N   : OUT std_logic;
    roD       : OUT std_logic_vector(31 DOWNTO 0);    
    roTEN_N   : OUT std_logic;
    roCMD_N   : OUT std_logic;
    riD       : IN  std_logic_vector(31 DOWNTO 0);
    riTEN_N   : IN  std_logic;
    riSTS_N   : IN  std_logic;

    -- Data output
    data      : OUT std_logic_vector(31 DOWNTO 0);
    d_valid   : OUT std_logic;
    eoe_flag  : OUT std_logic;
    busy_n    : IN  std_logic;
    
    -- config IF
    conf_addr   : IN  std_logic_vector(7 DOWNTO 0);
    conf_in     : IN  std_logic_vector(31 DOWNTO 0);
    conf_out    : OUT std_logic_vector(31 DOWNTO 0);
    conf_we     : IN  std_logic    
  );

END ddl_if;


ARCHITECTURE syn OF ddl_if IS

  SIGNAL we  : std_logic;

  SIGNAL riD_reg   : std_logic_vector(31 DOWNTO 0);
  SIGNAL riSTS_reg : std_logic;
  SIGNAL riTEN_reg : std_logic;

  SIGNAL ctstw_reg     : std_logic_vector(31 DOWNTO 0);
  SIGNAL festw_reg     : std_logic_vector(31 DOWNTO 0);
  SIGNAL dtstw_reg     : std_logic_vector(31 DOWNTO 0);
  SIGNAL ifstw_reg     : std_logic_vector(31 DOWNTO 0);
  SIGNAL stwrd_reg     : std_logic_vector(31 DOWNTO 0);
  SIGNAL commd_reg     : std_logic_vector(31 DOWNTO 0);

  SIGNAL startWatchDog : std_logic;
  SIGNAL abort_trans   : std_logic;
  SIGNAL start_trans   : std_logic;
  SIGNAL trans_count   : std_logic_vector(3 DOWNTO 0);
  
  TYPE trans_state_type IS (trans_idle, send_cmd, wait_ctstw);
  SIGNAL trans_state : trans_state_type;

  
  BEGIN

    roCLK   <= clk;
    roBSY_N <= busy_n;
    roRST_N <= rstn;
    
    start_trans <= we;

    -- End-Of-Event
    eoe_flag <= '1' WHEN (riTEN_reg = '0' AND
                          riSTS_reg = '0' AND
                          riD_reg(7 DOWNTO 0) = x"82") ELSE '0';

    data     <= (x"000" & '0' & stwrd_reg(30 DOWNTO 12)) WHEN (riTEN_reg = '0' AND
                                                               riSTS_reg = '0' AND
                                                               riD_reg(7 DOWNTO 0) = x"82")
                ELSE riD_reg;
    
    d_valid  <= '1' WHEN ((riTEN_reg='0' AND riSTS_reg='1') OR               -- normal data transfer
                          (riSTS_reg='0' AND riD_reg(7 DOWNTO 0)=x"82"))     -- DTSTW transfer
                   ELSE '0';
    
    ---------------------------------------------------------------------------
    config_if:
    ---------------------------------------------------------------------------
    we       <= '1' WHEN (conf_we='1' AND conf_addr = commd_reg_addr) ELSE '0';
    conf_out <= commd_reg WHEN (conf_addr = commd_reg_addr) ELSE
                ctstw_reg WHEN (conf_addr = ctstw_reg_addr) ELSE
                festw_reg WHEN (conf_addr = festw_reg_addr) ELSE
                dtstw_reg WHEN (conf_addr = dtstw_reg_addr) ELSE
                ifstw_reg WHEN (conf_addr = ifstw_reg_addr) ELSE
                (OTHERS => '0');

    -----------------------------------------------------------------------------
    ddl_input_reg:
    -----------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        riD_reg   <= (OTHERS => '0');
        riSTS_reg <= '1';
        riTEN_reg <= '1';
      ELSIF (clk'event and clk = '1') THEN
        riD_reg   <= riD;
        riSTS_reg <= riSTS_N;
        riTEN_reg <= riTEN_N;
      END IF;
    END PROCESS;

    
    ---------------------------------------------------------------------------
    command_register:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
      BEGIN
      IF (rstn = '0') THEN
        commd_reg <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (we = '1') THEN
          commd_reg <= conf_in;
        ELSE
          commd_reg <= commd_reg;
        END IF;
      END IF;  
    END PROCESS;
        
      
    -------------------------------------------------------------------------------
    WATCHDOG:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    VARIABLE count : std_logic_vector(4 downto 0);
    BEGIN 
      IF (rstn='0') THEN 
        count := "00000";
      ELSIF ( clk'event AND clk='1') THEN
        IF (startWatchDog='1') THEN 
          count := "00000";
        ELSE
          count := count + '1';
        END IF ;
      END IF;
    
      abort_trans <= count(4);
    END PROCESS;

    
    -----------------------------------------------------------------------------
    TRANSACTION_COUNT:
    -----------------------------------------------------------------------------
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          trans_count <= (OTHERS => '0');
        ELSIF (clk'event AND clk='1') THEN
          IF (start_trans='1') THEN
            trans_count <= trans_count + 1;
          ELSE
            trans_count <= trans_count;
          END IF;
        ELSE
          trans_count <= trans_count;
        END IF;
      END PROCESS;


    -------------------------------------------------------------------------
    catch_status_word:
    -------------------------------------------------------------------------
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn = '0') THEN
          stwrd_reg <= (OTHERS => '1');
        ELSIF (clk'event AND clk = '1') THEN
          IF (start_trans='1') THEN
            stwrd_reg <= (OTHERS => '1');
          ELSIF (riSTS_N='0' AND riTEN_N='0') THEN
            stwrd_reg <= riD;
          ELSE
            stwrd_reg <= stwrd_reg;
          END IF;
        END IF;
      END PROCESS;

    PROCESS (clk, rstn)
      BEGIN
        IF (rstn = '0') THEN
          ctstw_reg <= (OTHERS => '1');
          festw_reg <= (OTHERS => '1');
          dtstw_reg <= (OTHERS => '1');
          ifstw_reg <= (OTHERS => '1');
        ELSIF (clk'event AND clk = '1') THEN
          
          IF (trans_state = send_cmd) THEN
            ctstw_reg <= (OTHERS => '1');
          ELSIF (stwrd_reg(7 DOWNTO 6) = "00") THEN
            ctstw_reg <= stwrd_reg;
          ELSE
            ctstw_reg <= ctstw_reg;
          END IF;

          IF (trans_state = send_cmd AND (commd_reg(7 DOWNTO 0)=x"14" OR commd_reg(7 DOWNTO 0)=x"b4")) THEN
            festw_reg <= (OTHERS => '1');
          ELSIF (stwrd_reg(7 DOWNTO 6) = "01") THEN
            festw_reg <= stwrd_reg;
          ELSE
            festw_reg <= festw_reg;
          END IF;

          IF (trans_state = send_cmd AND commd_reg(7 DOWNTO 0) = x"14") THEN
            dtstw_reg <= (OTHERS => '1');
          ELSIF (stwrd_reg(7 DOWNTO 6) = "10") THEN
            dtstw_reg <= stwrd_reg;
          ELSE
            dtstw_reg <= dtstw_reg;
          END IF;

          IF (trans_state = send_cmd) THEN
            ifstw_reg <= (OTHERS => '1');
          ELSIF (stwrd_reg(7 DOWNTO 6) = "11") THEN
            ifstw_reg <= stwrd_reg;
          ELSE
            ifstw_reg <= ifstw_reg;
          END IF;
          
        END IF;
      END PROCESS;





  -----------------------------------------------------------------------------
  trans_controller:
  -----------------------------------------------------------------------------
  PROCESS (clk, rstn)
  BEGIN
    IF (rstn='0') THEN
      trans_state <= trans_idle;
    ELSIF (clk'event AND clk='1') THEN
      CASE trans_state IS
        WHEN trans_idle =>
          IF (start_trans = '1') THEN
            trans_state <= send_cmd;
          ELSE
            trans_state <= trans_idle;
          END IF;
          
        WHEN send_cmd =>
          trans_state <= wait_ctstw;
          
        WHEN wait_ctstw =>
          IF (abort_trans = '1') THEN
            trans_state <= trans_idle;
          ELSIF (stwrd_reg(7 DOWNTO 6) = "00") THEN
            trans_state <= trans_idle;
          ELSE
            trans_state <= wait_ctstw;
          END IF;

        WHEN OTHERS =>
            trans_state <= trans_idle;

      END CASE;
    END IF;
  END PROCESS;

  -----------------------------------------------------------------------------
  trans_controller_data:
  -----------------------------------------------------------------------------    
  PROCESS (trans_state, riTEN_N, trans_count, commd_reg)
  BEGIN
    -- default values
    roTEN_N <= '1';
    roCMD_N <= '1';
    roD     <= (OTHERS => '0');
    startWatchDog <= '0';
    
    CASE trans_state IS

      WHEN trans_idle =>
        startWatchDog <= '0';
        
      WHEN send_cmd =>
        roCMD_N <= '0';
        roTEN_N <= '0';
        roD     <= x"00000" & trans_count & commd_reg(7 DOWNTO 0);
        startWatchDog <= '1';
        
      WHEN wait_ctstw =>
        startWatchDog <= '0';
        
      WHEN OTHERS =>

    END CASE;
  END PROCESS;

END syn;


