--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : local_side.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de>
-- Company    : 
-- Platform   : 
-------------------------------------------------------------------------------
-- Description:
--
-- Stores DMA descriptors for each DMA transfer.
-- Initiate DMA transfers by starting "master_ctrl" via signals
-- "start_mstr" and "mstr_done".
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author    Description
--
-- 07.07.2004  1.0      tilsner   created
--=============================================================================

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;

ENTITY dma_engine IS
  GENERIC (
    dma_wrptr      : integer RANGE 0 TO 255 := 0;
    dma_buflen     : integer RANGE 0 TO 255 := 8;
    dma_rptptr     : integer RANGE 0 TO 255 := 12;
    dma_rptptr_len : integer RANGE 0 TO 255 := 20;
    dma_status     : integer RANGE 0 TO 255 := 16
  );
  PORT (
    --system
    clk          : IN  std_logic;
    rstn         : IN  std_logic;
    mstr_done    : IN  std_logic;
    start_mstr   : OUT std_logic;
    dma_addr     : OUT std_logic_vector(31 DOWNTO 0);
    rpt_addr     : OUT std_logic_vector(31 DOWNTO 0);
    rd_ptr       : IN  std_logic_vector(31 DOWNTO 0);
    dma_en       : IN  std_logic;
    dma_busy     : OUT std_logic;
    fifo_usedw   : IN  std_logic_vector(7 downto 0);
    fifo_full    : IN  std_logic;
    burst_no     : OUT std_logic_vector(31 DOWNTO 0);
    last_block   : IN  std_logic;
    hostbuf_full : OUT std_logic;

    -- config IF
    conf_addr    : IN  std_logic_vector(7 DOWNTO 0);
    conf_in      : IN  std_logic_vector(31 DOWNTO 0);
    conf_out     : OUT std_logic_vector(31 DOWNTO 0);
    conf_we      : IN  std_logic
  );

END dma_engine;



ARCHITECTURE behaviour OF dma_engine IS
  
  -- state machine encoding
  TYPE trig_state_type IS (dma_idle, dma_triggered);
  SIGNAL dma_trig_state : trig_state_type;

  SIGNAL we  : std_logic_vector(3 DOWNTO 0);
  
  SIGNAL base_addr        : std_logic_vector(31 DOWNTO 0);
  SIGNAL block_len        : std_logic_vector(31 DOWNTO 0);
  SIGNAL report_base      : std_logic_vector(31 DOWNTO 0);
  SIGNAL report_len       : std_logic_vector(31 DOWNTO 0);
  SIGNAL dma_addr_inc     : std_logic_vector(31 DOWNTO 0);
  SIGNAL report_addr_inc  : std_logic_vector(31 DOWNTO 0);
  SIGNAL burst_count      : std_logic_vector(31 DOWNTO 0);
  SIGNAL tot_burst_count  : std_logic_vector(31 DOWNTO 0);
  SIGNAL start_dma_tx     : std_logic;
  SIGNAL ed_start_mstr    : std_logic;
  SIGNAL inc_burst_cnt    : std_logic;
  SIGNAL output_sig       : std_logic_vector(31 DOWNTO 0);
  SIGNAL start_mstr_int   : std_logic;
  SIGNAL hostbuf_full_int : std_logic;
  
  SIGNAL buffer_full      : std_logic;
  SIGNAL buffer_max       : std_logic_vector(31 DOWNTO 0);
  SIGNAL report_max       : std_logic_vector(31 DOWNTO 0);
  SIGNAL wr_offset        : std_logic_vector(31 DOWNTO 0);
  SIGNAL rd_offset        : std_logic_vector(31 DOWNTO 0);
  SIGNAL max_offset       : std_logic_vector(31 DOWNTO 0);
  
  BEGIN

    dma_busy     <= '1' WHEN (dma_trig_state = dma_triggered) ELSE '0';
    burst_no     <= burst_count;
    start_dma_tx <= '1' WHEN (dma_en='1' AND hostbuf_full_int='0'
                             AND (fifo_usedw > 15 OR fifo_full='1'))
                    ELSE '0';

    hostbuf_full     <= buffer_full;    -- output port
    hostbuf_full_int <= buffer_full;    -- internal signal

    wr_offset    <= dma_addr_inc - base_addr;
    rd_offset    <= rd_ptr - base_addr;
    max_offset   <= block_len - 128;
    buffer_max   <= base_addr + block_len - 128;
    report_max   <= report_base + report_len - 4;
    buffer_full  <= '1' WHEN ((wr_offset+128 = rd_offset) OR                  -- normal case
                              ((wr_offset=max_offset) AND (rd_ptr=base_addr))) -- buffer wrap
                    ELSE '0';

    ---------------------------------------------------------------------------
    config_if:
    ---------------------------------------------------------------------------
    we(0)    <= '1' WHEN (conf_we='1' AND conf_addr = dma_wrptr)  ELSE '0';
    we(1)    <= '1' WHEN (conf_we='1' AND conf_addr = dma_buflen) ELSE '0';
    we(2)    <= '1' WHEN (conf_we='1' AND conf_addr = dma_rptptr) ELSE '0';
    we(3)    <= '1' WHEN (conf_we='1' AND conf_addr = dma_rptptr_len) ELSE '0';
    conf_out <= dma_addr_inc    WHEN (conf_addr = dma_wrptr) ELSE
                block_len       WHEN (conf_addr = dma_buflen) ELSE
                report_addr_inc WHEN (conf_addr = dma_rptptr) ELSE
                report_len      WHEN (conf_addr = dma_rptptr_len) ELSE
                output_sig      WHEN (conf_addr = dma_status) ELSE
                (OTHERS => '0');

    output_sig <=  "000" & x"0000" & burst_count(12 DOWNTO 0);

   

    -------------------------------------------------------------------------------
    ctrl_register:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        base_addr   <= (OTHERS => '0');
        block_len   <= (OTHERS => '0');
        report_base <= (OTHERS => '0');
        report_len  <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (we(0) = '1') THEN
          base_addr <= conf_in;
        END IF;
        IF (we(1) = '1') THEN
          block_len <= conf_in;
        END IF;
        IF (we(2) = '1') THEN
          report_base <= conf_in;
        END IF;
        IF (we(3) = '1') THEN
          report_len <= conf_in;
        END IF;
      END IF;  
    END PROCESS;

    -------------------------------------------------------------------------------
    dma_trigger:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        dma_trig_state  <= dma_idle;
      ELSIF (clk'event and clk = '1') THEN
        CASE dma_trig_state IS
          WHEN dma_idle =>
            IF (start_dma_tx='1') THEN
              dma_trig_state <= dma_triggered;
            ELSE
              dma_trig_state <= dma_idle;
            END IF;
            
          WHEN dma_triggered =>
            IF (mstr_done = '1') THEN
              dma_trig_state <= dma_idle;
            ELSE 
              dma_trig_state <= dma_triggered;
            END IF;
              
          WHEN OTHERS =>
            dma_trig_state <= dma_idle;
            
        END CASE;
      END IF;
    END PROCESS;

    start_mstr     <= '1' WHEN (dma_trig_state = dma_triggered) ELSE '0';
    start_mstr_int <= '1' WHEN (dma_trig_state = dma_triggered) ELSE '0';
      
    -------------------------------------------------------------------------------
    dma_address_increment:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        dma_addr_inc <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (we(0) = '1') THEN
          dma_addr_inc <= conf_in;
        ELSIF ((dma_addr_inc = buffer_max) AND (mstr_done='1')) THEN
          dma_addr_inc <= base_addr;
        ELSIF (mstr_done='1') THEN
          dma_addr_inc <= dma_addr_inc + x"80";
        ELSE
          dma_addr_inc <= dma_addr_inc;
        END IF;
      END IF;
    END PROCESS;


    -------------------------------------------------------------------------------
    report_address_increment:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        report_addr_inc <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (we(2) = '1') THEN
          report_addr_inc <= conf_in;
        ELSIF (report_addr_inc = report_max AND mstr_done='1' AND last_block='1') THEN
          report_addr_inc <= report_base;
        ELSIF (mstr_done='1' AND last_block='1') THEN
          report_addr_inc <= report_addr_inc + x"4";
        ELSE
          report_addr_inc <= report_addr_inc;
        END IF;
      END IF;
    END PROCESS;

    dma_addr <= dma_addr_inc;
    rpt_addr <= report_addr_inc;
    
    -------------------------------------------------------------------------------
    burst_counter:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        burst_count <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (mstr_done='1' AND last_block='1') THEN
          burst_count <= (OTHERS => '0');
        ELSIF (inc_burst_cnt = '1') THEN
          burst_count <= burst_count + '1';
        END IF;
      END IF;
    END PROCESS;
    
    -------------------------------------------------------------------------------
    total_burst_counter:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        tot_burst_count <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (we(2)='1') THEN
          tot_burst_count <= (OTHERS => '0');
        ELSIF (inc_burst_cnt = '1') THEN
          tot_burst_count <= tot_burst_count + '1';
        END IF;
      END IF;
    END PROCESS;

    -------------------------------------------------------------------------------
    edge_detect_start_mstr:
    -------------------------------------------------------------------------------    
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        ed_start_mstr <= '0';
      ELSIF (clk'event and clk = '1') THEN
        ed_start_mstr <= start_mstr_int;
      END IF;
    END PROCESS;
    inc_burst_cnt <= '1' WHEN (ed_start_mstr='0' AND start_mstr_int='1') ELSE '0';

    
  END behaviour;
