--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : dma_fifo.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de> / Olav Torheim <torheim@ift.uib.no>
-- Company    :
-- Platform   :
-------------------------------------------------------------------------------
-- Description:
--
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author    Description
--
-- 07.07.2004  1.0      tilsner   created
-- 24.06.2005  1.1      O. Torheim  Replacing the Altera-based FIFO interface with Xilinx-based FIFO and adding some glue logic
--=============================================================================

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_signed.ALL;


ENTITY dma_fifo IS

  PORT (
    clk        : IN  STD_LOGIC ;
    clk_global : IN  STD_LOGIC ;
    rstn       : IN  STD_LOGIC;
    data       : IN  std_logic_vector(31 DOWNTO 0);
    d_valid    : IN  STD_LOGIC;
    eoe_flag   : IN  STD_LOGIC;
    fifo_rd    : IN  STD_LOGIC;
    fifo_clear : IN  STD_LOGIC;
    fifo_out   : OUT STD_LOGIC_VECTOR(64 DOWNTO 0);
    full       : OUT STD_LOGIC;
    empty      : INOUT STD_LOGIC;
    usedw      : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0);
    event_length : OUT STD_LOGIC_VECTOR(31 DOWNTO 0)
  );    

END dma_fifo;


ARCHITECTURE rtl OF dma_fifo IS

  TYPE prefetch_state_type IS (fifo_empty, fifo_filled, fifo_growing, fifo_idle); -- OT 26.05.05

  SIGNAL prefetch_state : prefetch_state_type;
  SIGNAL former_prefetch_state : prefetch_state_type;

    SIGNAL data_reg       : std_logic_vector(31 DOWNTO 0);
    SIGNAL d_valid_reg    : STD_LOGIC;
    SIGNAL eoe_flag_reg   : STD_LOGIC;

--  COMPONENT fifo_256x65bit
--    PORT (
--      data  : IN STD_LOGIC_VECTOR(64 DOWNTO 0);
--      wrreq : IN STD_LOGIC ;
--      rdreq : IN STD_LOGIC ;
--      clock : IN STD_LOGIC ;
--      aclr  : IN STD_LOGIC ;
--      q	    : OUT STD_LOGIC_VECTOR(64 DOWNTO 0);
--      full  : OUT STD_LOGIC ;
--      empty : OUT STD_LOGIC ;
--      usedw : OUT STD_LOGIC_VECTOR(7 DOWNTO 0)     
--    );
--  END COMPONENT;

  COMPONENT fifo_256x65bit
	port (
	clk: IN std_logic;
	din: IN std_logic_VECTOR(64 downto 0);
	rd_en: IN std_logic;
	rst: IN std_logic;
	wr_en: IN std_logic;
	data_count: OUT std_logic_VECTOR(7 downto 0);
	dout: OUT std_logic_VECTOR(64 downto 0);
	empty: OUT std_logic;
	full: OUT std_logic
    );
  END COMPONENT;	 

  SIGNAL fi_64     : STD_LOGIC_VECTOR(64 DOWNTO 0);
  SIGNAL fifo_lw   : std_logic_vector(31 DOWNTO 0);  -- lower word of fifo input
  SIGNAL ff_en     : std_logic;
  SIGNAL fifo_q    : std_logic_vector(64 DOWNTO 0);
  SIGNAL wrreq     : std_logic;
  SIGNAL aclr      : std_logic;
  SIGNAL sel       : std_logic_vector(1 DOWNTO 0);
  
  SIGNAL rst		   : std_logic; -- OT 26.05.05

  SIGNAL fifo_rd_data      : std_logic; -- OT 26.06.05

BEGIN  -- rtl

  aclr     <= NOT(rstn) OR fifo_clear;
  fi_64    <= eoe_flag_reg & data_reg & fifo_lw;
  fifo_out <= fifo_q;
  wrreq    <= ((NOT ff_en) AND d_valid_reg) OR eoe_flag_reg;

--  fifo: fifo_256x65bit
--    PORT MAP (
--      data  => fi_64,
--      wrreq => wrreq,
--      rdreq => fifo_rd,
--      clock => clk,
--      aclr  => aclr,
--      q	    => fifo_q,
--      full  => full,
--      empty => empty,
--      usedw => usedw
--    );

  fifo: fifo_256x65bit
    PORT MAP (
	clk => clk,
	din => fi_64,
	rd_en => fifo_rd_data,
	rst => rst,
	wr_en => wrreq,
	data_count => usedw,
	dout => fifo_q,
	empty => empty,
	full => full
    );

  -----------------------------------------------------------------------------
  -- FIFO input
  -- combines two 32bit words into one 64bit word
  -----------------------------------------------------------------------------
  PROCESS (clk, rstn)
    BEGIN
      IF (rstn='0') THEN
        fifo_lw <= (OTHERS => '0');
      ELSIF (clk'event AND clk='1') THEN
        IF (ff_en = '1') THEN
          fifo_lw <= data_reg;
        END IF;
      END IF;

    END PROCESS;

  -----------------------------------------------------------------------------
  -- Toggle FF
  -- controls the concatenation of 32bit words
  -----------------------------------------------------------------------------
  PROCESS (clk, rstn, d_valid_reg, eoe_flag_reg)
    BEGIN
      IF (rstn='0') THEN
        ff_en <= '1';
      ELSIF (clk'event AND clk='1') THEN
        IF (eoe_flag_reg = '1') THEN
          ff_en <= '1';
        ELSIF (d_valid_reg = '1') THEN
          ff_en <= NOT ff_en;
        END IF;
      END IF;
    END PROCESS;


  -----------------------------------------------------------------------------
  -- event length register
  -----------------------------------------------------------------------------
  PROCESS (clk, rstn)
  BEGIN
    IF (rstn = '0') THEN
      event_length <= (OTHERS => '0');
    ELSIF rising_edge(clk) THEN
      IF (eoe_flag_reg = '1') THEN
        event_length <= data_reg(31 DOWNTO 0);
      END IF;
    END IF;
  END PROCESS;

	rst <= not rstn;

  -----------------------------------------------------------------------------
  -- Prefetch first FIFO element
  -----------------------------------------------------------------------------
  PROCESS (clk, rstn)
  BEGIN
    IF (rstn = '0') THEN
       prefetch_state <= fifo_empty;
       former_prefetch_state <= fifo_empty;
    ELSIF rising_edge(clk) THEN
          former_prefetch_state <= prefetch_state;
        CASE prefetch_state IS

             WHEN fifo_empty =>
                  IF (empty='0') THEN
                     prefetch_state <= fifo_growing;
                  ELSE
                      prefetch_state <= fifo_empty;
                  END IF;

             WHEN fifo_filled =>
                  IF (usedw=15) THEN
                     prefetch_state <= fifo_idle;
                  ELSE
                      prefetch_state <= fifo_filled;
                  END IF;

             WHEN fifo_growing =>
                  IF (usedw=16) THEN
                     prefetch_state <= fifo_filled;
                  ELSE
                      prefetch_state <= fifo_growing;
                  END IF;

             WHEN fifo_idle =>
                      prefetch_state <= fifo_idle;
             WHEN others =>
                  prefetch_state <= fifo_empty;
        END CASE;
    END IF;
  END PROCESS;

    fifo_rd_data <= '1' WHEN (fifo_rd='1' OR
                         (prefetch_state = fifo_filled AND former_prefetch_state=fifo_growing)) ELSE '0';

  -----------------------------------------------------------------------------
  -- Keep it registered
  -----------------------------------------------------------------------------
  PROCESS (clk_global, rstn)
  BEGIN
    IF (rstn = '0') THEN
    data_reg <= (OTHERS => '0') ;
     d_valid_reg    <= '0';
     eoe_flag_reg   <= '0';
    ELSIF rising_edge(clk_global) THEN
     data_reg       <= data;
     d_valid_reg    <= d_valid;
     eoe_flag_reg   <= eoe_flag;
    END IF;
  END PROCESS;
END rtl;

