--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : local_side.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de> / Artur Szostak / Olav Torheim <torheim@ift.uib.no>
-- Company    :
-- Platform   :
-------------------------------------------------------------------------------
-- Description:
--
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author      Description
--
-- 07.07.2004  1.0      tilsner     created
--
-- 09.03.2005  1.1      A. Szostak  Binding register addresses to those defined
--                                  in HLT_RORC_definitions.vhd
--
-- 24.06.2005  1.2      O. Torheim  Replacing the Altera MegaCore interface with the Xilinx LogiCore interface
--                                  Adding error handling (latency timer expiration and retries on the PCI bus)
--=============================================================================

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_signed.ALL;
USE work.HLT_RORC_definitions.ALL;

ENTITY local_side IS
  PORT (
    -- pci signals
        FRAMEQ_N        : in    std_logic;
        REQ64Q_N        : in    std_logic;
        TRDYQ_N         : in    std_logic;
        IRDYQ_N         : in    std_logic;
        STOPQ_N         : in    std_logic;
        DEVSELQ_N       : in    std_logic;
        ACK64Q_N        : in    std_logic;
        ADDR            : in    std_logic_vector( 31 downto 0);
        ADIO            : inout std_logic_vector( 63 downto 0);
        CFG_VLD         : in    std_logic;
        CFG_HIT         : in    std_logic;
        C_TERM          : out   std_logic;
        C_READY         : out   std_logic;
        ADDR_VLD        : in    std_logic;
        BASE_HIT        : in    std_logic_vector(  7 downto 0);
        S_CYCLE64       : in    std_logic;
        S_TERM          : out   std_logic;
        S_READY         : out   std_logic;
        S_ABORT         : out   std_logic;
        S_WRDN          : in    std_logic;
        S_SRC_EN        : in    std_logic;
        S_DATA_VLD      : in    std_logic;
        S_CBE           : in    std_logic_vector(  7 downto 0);
        PCI_CMD         : in    std_logic_vector( 15 downto 0);
        REQUEST         : out   std_logic;
        REQUEST64       : out   std_logic;
        REQUESTHOLD     : out   std_logic;
        COMPLETE        : out   std_logic;
        M_WRDN          : out   std_logic;
        M_READY         : out   std_logic;
        M_SRC_EN        : in    std_logic;
        M_DATA_VLD      : in    std_logic;
        M_CBE           : out   std_logic_vector(  7 downto 0);
        TIME_OUT        : in    std_logic;
        M_FAIL64        : in    std_logic;
        CFG_SELF        : out   std_logic;
        M_DATA          : in    std_logic;
        DR_BUS          : in    std_logic;
        I_IDLE          : in    std_logic;
        M_ADDR_N        : in    std_logic;
        IDLE            : in    std_logic;
        B_BUSY          : in    std_logic;
        S_DATA          : in    std_logic;
        BACKOFF         : in    std_logic;
        SLOT64          : out   std_logic;
        INTR_N          : out   std_logic; -- interrupt
        PERRQ_N         : in    std_logic;
        SERRQ_N         : in    std_logic;
        KEEPOUT         : out   std_logic;
        CSR             : in    std_logic_vector( 39 downto 0);
        SUB_DATA        : out   std_logic_vector( 31 downto 0);
        CFG             : in    std_logic_vector(255 downto 0);
        RST             : in    std_logic;
        CLK             : in    std_logic;
        CLK_REG         : in    std_logic;                

    -- DDL interface
    roCLK           : OUT std_logic;
    roRST_N         : OUT std_logic;
    riLD_N          : IN  std_logic;
    riLF_N          : IN  std_logic;
    roBSY_N         : OUT std_logic;
    roD             : OUT std_logic_vector(31 DOWNTO 0);
    roTEN_N         : OUT std_logic;
    roCMD_N         : OUT std_logic;
    riD             : IN  std_logic_vector(31 DOWNTO 0);
    riTEN_N         : IN  std_logic;
    riSTS_N         : IN  std_logic;

    -- SRAM32 interaface
    sram32_a       : OUT   std_logic_vector(23 DOWNTO 0);
    sram32_d       : INOUT std_logic_vector(31 DOWNTO 0);
    sram32_we      : OUT   std_logic;
    sram32_be      : OUT   std_logic_vector(3 DOWNTO 0);
    sram32_oe      : OUT   std_logic;

    -- SRAM64 interface
    sram64_a       : OUT   std_logic_vector(21 DOWNTO 0);
    sram64_d       : INOUT std_logic_vector(63 DOWNTO 0);
    sram64_we      : OUT   std_logic_vector(1 DOWNTO 0);
    sram64_be      : OUT   std_logic_vector(7 DOWNTO 0);
    sram64_oe      : OUT   std_logic_vector(1 DOWNTO 0);
    sram64_cs      : OUT   std_logic;

    -- status
    hostbuf_full    : OUT std_logic;
    dmabuf_full     : OUT std_logic;
    dma_enabled     : OUT std_logic;
    dma_active      : OUT std_logic;
    link_down       : OUT std_logic
    );

END local_side;


ARCHITECTURE rtl OF local_side IS

  COMPONENT csr_block
    GENERIC (
      global_csr : integer RANGE 0 TO 255;
      global_rst : integer RANGE 0 TO 255;
      fclear_cmd : integer RANGE 0 TO 255
    );
    PORT (
      clk         : IN  std_logic;
      rstn        : IN  std_logic;
      s_rst       : OUT std_logic;
      busy_en     : OUT std_logic;
      pg_en       : OUT std_logic;
      ram_en      : OUT std_logic;
      ddl_en      : OUT std_logic;
      dma_en      : OUT std_logic;
      irq_en		  : OUT std_logic;
      fifo_clear  : OUT std_logic;
      dma_busy    : IN  std_logic;
      fifo_usedw  : IN  std_logic_vector(7 DOWNTO 0);
      fifo_full   : IN  std_logic;
      fifo_empty  : IN  std_logic;
      hostbuf_full: IN  std_logic;
      pci_64bit   : IN  std_logic;
      conf_addr   : IN  std_logic_vector(7 DOWNTO 0);
      conf_in     : IN  std_logic_vector(31 DOWNTO 0);
      conf_out    : OUT std_logic_vector(31 DOWNTO 0);
      conf_we     : IN  std_logic
    );
  END COMPONENT;

  COMPONENT target_ctrl
    PORT (
      clk           : IN  std_logic;
      rstn          : IN  std_logic;
        ADDR            : in    std_logic_vector( 31 downto 0);
        ADIO            : inout std_logic_vector( 63 downto 0);
        BASE_HIT        : in    std_logic_vector(  7 downto 0);
        S_ABORT         : out   std_logic;
        S_WRDN          : in    std_logic;
        IDLE            : in    std_logic;
        S_DATA_VLD      : in    std_logic;
      pci_out       : OUT std_logic;
      conf_rd       : IN  std_logic_vector(31 DOWNTO 0);
      conf_wr       : OUT std_logic_vector(31 DOWNTO 0);
      conf_addr     : OUT std_logic_vector(7 DOWNTO 0);
      conf_we       : OUT std_logic
    );
  END COMPONENT;

  COMPONENT master_ctrl
    GENERIC (
      rd_ptr_addr_reg_addr  : integer RANGE 0 TO 255;
      rd_ptr_reg_addr       : integer RANGE 0 TO 255;
      dma_error_reg_addr    : integer RANGE 0 TO 255;
      report_error_reg_addr : integer RANGE 0 TO 255;
      read_error_reg_addr   : integer RANGE 0 TO 255
    );
    PORT (
      clk          : in  std_logic;
      rstn         : in  std_logic;
    ADIO            : inout std_logic_vector( 63 downto 0);
    CSR             : in    std_logic_vector( 39 downto 0);
    M_DATA          : in    std_logic;
    M_DATA_VLD      : in    std_logic;
    M_ADDR_N        : in    std_logic;
    M_CBE           : out   std_logic_vector(  7 downto 0);
    M_WRDN          : out   std_logic;
    M_SRC_EN        : in    std_logic;
    M_FAIL64        : in    std_logic;
    TIME_OUT        : in    std_logic;
    COMPLETE        : out   std_logic;
    REQUEST         : out   std_logic;
    REQUEST64       : out   std_logic;
    REQUESTHOLD     : out   std_logic;

      stat_reg     : IN  std_logic_vector(5 DOWNTO 0);
      dma_addr     : IN  std_logic_vector(31 downto 0);  -- dma target address
      rpt_addr     : IN  std_logic_vector(31 DOWNTO 0);  -- dma report address
      rd_ptr       : OUT std_logic_vector(31 DOWNTO 0);
      start_mstr   : IN  std_logic;                      -- signal starts master
      mstr_done    : OUT std_logic;                      -- signal to DMA engine, master finished
      last_block   : IN  std_logic;
      hostbuf_full : IN  std_logic;
      burst_count  : IN  std_logic_vector(31 DOWNTO 0);
      event_length : IN  std_logic_vector(18 DOWNTO 0);
      pci_out      : OUT std_logic;
      oe_io_reg    : INOUT std_logic;
      fifo_rd      : OUT std_logic;
      burst_cont   : OUT std_logic;
      irq_en       : IN std_logic; -- OT 18.02.05
      irq_generate : OUT std_logic; -- OT 18.02.05
      dma_buffer_full : out std_logic;
      end_of_event    : out std_logic;
      conf_addr    : IN  std_logic_vector(7 DOWNTO 0);
      conf_in      : IN  std_logic_vector(31 DOWNTO 0);
      conf_out     : OUT std_logic_vector(31 DOWNTO 0);
      conf_we      : IN  std_logic
    );
  END COMPONENT;
  
  COMPONENT dma_engine
    GENERIC (
      dma_wrptr      : integer RANGE 0 TO 255;
      dma_buflen     : integer RANGE 0 TO 255;
      dma_rptptr     : integer RANGE 0 TO 255;
      dma_rptptr_len : integer RANGE 0 TO 255;
      dma_status     : integer RANGE 0 TO 255
    );
    PORT (
      clk          : IN  std_logic;
      rstn         : IN  std_logic;
      mstr_done    : IN  std_logic;
      start_mstr   : OUT std_logic;
      dma_addr     : OUT std_logic_vector(31 DOWNTO 0);
      rpt_addr     : OUT std_logic_vector(31 DOWNTO 0);
      rd_ptr       : IN  std_logic_vector(31 DOWNTO 0);
      dma_en       : IN  std_logic;
      dma_busy     : OUT std_logic;
      fifo_usedw   : IN  std_logic_vector(7 DOWNTO 0);
      fifo_full    : IN  std_logic;
      burst_no     : OUT std_logic_vector(31 DOWNTO 0);
      last_block   : IN  std_logic;
      hostbuf_full : OUT std_logic;
      conf_addr    : IN  std_logic_vector(7 DOWNTO 0);
      conf_in      : IN  std_logic_vector(31 DOWNTO 0);
      conf_out     : OUT std_logic_vector(31 DOWNTO 0);
      conf_we      : IN  std_logic      
    );
  END COMPONENT;

  COMPONENT pattern_generator
    GENERIC (
      patgen_evlen_reg : integer RANGE 0 TO 255;
      patgen_csr_reg   : integer RANGE 0 TO 255;
      patgen_start_reg : integer RANGE 0 TO 255
    );
    PORT (
      clk       : IN  std_logic;
      rstn      : IN  std_logic;
      data      : OUT std_logic_vector(31 DOWNTO 0);
      d_valid   : OUT std_logic;
      eoe_flag  : OUT std_logic;
      busy_n    : IN  std_logic;
      conf_addr : IN  std_logic_vector(7 DOWNTO 0);
      conf_in   : IN  std_logic_vector(31 DOWNTO 0);
      conf_out  : OUT std_logic_vector(31 DOWNTO 0);
      conf_we   : IN  std_logic
    );
  END COMPONENT;

  COMPONENT dma_fifo
    PORT (
      clk          : IN  STD_LOGIC ;
    clk_global : IN  STD_LOGIC ;      
      rstn         : IN  STD_LOGIC;
      data         : IN  std_logic_vector(31 DOWNTO 0);
      d_valid      : IN  STD_LOGIC;
      eoe_flag     : IN  STD_LOGIC;
      fifo_rd      : IN  STD_LOGIC;
      fifo_clear   : IN  STD_LOGIC;
      fifo_out     : OUT STD_LOGIC_VECTOR (64 DOWNTO 0);
      full         : OUT STD_LOGIC ;
      empty        : INOUT STD_LOGIC ;
      usedw        : INOUT STD_LOGIC_VECTOR (7 DOWNTO 0);
      event_length : OUT STD_LOGIC_VECTOR(31 DOWNTO 0)
    );
  END COMPONENT;

  COMPONENT ddl_if
    GENERIC (
      commd_reg_addr : integer RANGE 0 TO 255;
      ctstw_reg_addr : integer RANGE 0 TO 255;
      festw_reg_addr : integer RANGE 0 TO 255;
      dtstw_reg_addr : integer RANGE 0 TO 255;
      ifstw_reg_addr : integer RANGE 0 TO 255
    );
    PORT (
      clk          : IN  std_logic;
      rstn         : IN  std_logic;
      roCLK        : OUT std_logic;
      roRST_N      : OUT std_logic;
      riLD_N       : IN  std_logic;
      riLF_N       : IN  std_logic;
      roBSY_N      : OUT std_logic;
      roD          : OUT std_logic_vector(31 DOWNTO 0);
      roTEN_N      : OUT std_logic;
      roCMD_N      : OUT std_logic;
      riD          : IN  std_logic_vector(31 DOWNTO 0);
      riTEN_N      : IN  std_logic;
      riSTS_N      : IN  std_logic;
      data         : OUT std_logic_vector(31 DOWNTO 0);
      d_valid      : OUT std_logic;
      eoe_flag     : OUT std_logic;
      busy_n       : IN  std_logic;
      conf_addr    : IN  std_logic_vector(7 DOWNTO 0);
      conf_in      : IN  std_logic_vector(31 DOWNTO 0);
      conf_out     : OUT std_logic_vector(31 DOWNTO 0);
      conf_we      : IN  std_logic
    );
  END COMPONENT;

  COMPONENT mem_if
    GENERIC (
      memif_addr    : integer RANGE 0 TO 255;
      memif_length  : integer RANGE 0 TO 255;
      memif_csr     : integer RANGE 0 TO 255;
      memif_trigger : integer RANGE 0 TO 255);
    PORT (
      clk       : IN  std_logic;
      rstn      : IN  std_logic;
      addr      : OUT std_logic_vector(23 DOWNTO 0);
      d         : IN  std_logic_vector(31 DOWNTO 0);
      oe        : OUT std_logic;
      data      : OUT std_logic_vector(31 DOWNTO 0);
      d_valid   : OUT std_logic;
      eoe_flag  : OUT std_logic;
      busy_n    : IN  std_logic;
      active    : OUT std_logic;
      conf_addr : IN  std_logic_vector(7 DOWNTO 0);
      conf_in   : IN  std_logic_vector(31 DOWNTO 0);
      conf_out  : OUT std_logic_vector(31 DOWNTO 0);
      conf_we   : IN  std_logic);
  END COMPONENT;

  COMPONENT packetizer
    PORT (
      clk        : IN  std_logic;
      rstn       : IN  std_logic;
      rdreq_in   : IN  std_logic;
      rdreq_out  : OUT std_logic;
      burst_continue   : IN std_logic;
      mstr_done  : IN  std_logic;
      last_block : OUT std_logic;
      data_in    : IN  std_logic_vector(64 DOWNTO 0);
      data_out   : OUT std_logic_vector(63 DOWNTO 0)
    );
  END COMPONENT;

  COMPONENT event_monitor
    GENERIC (
      evmon_counter : integer RANGE 0 TO 255
    );
    PORT (
      clk          : IN  std_logic;
      rstn         : IN  std_logic;
      data         : IN  std_logic_vector(64 DOWNTO 0);
      eoe_count    : OUT std_logic_vector(7 DOWNTO 0);
      event_count  : OUT std_logic_vector(15 DOWNTO 0);
      conf_addr    : IN  std_logic_vector(7 DOWNTO 0);
      conf_in      : IN  std_logic_vector(31 DOWNTO 0);
      conf_out     : OUT std_logic_vector(31 DOWNTO 0);
      conf_we      : IN  std_logic      
    );
  END COMPONENT;

  COMPONENT pci2mem
    GENERIC (
      bar_mem32 : integer RANGE 0 TO 15;
      bar_mem64 : integer RANGE 0 TO 15);
    PORT (
      clk         : in  std_logic;
      rstn        : in  std_logic;
      l_adro      : in  std_logic_vector(63 downto 0);
      l_cmdo      : in  std_logic_vector(3 downto 0);
      l_ldat_ackn : in  std_logic;
      l_hdat_ackn : in  std_logic;
      lt_abortn   : out std_logic;
      lt_discn    : out std_logic;
      lt_rdyn     : out std_logic;
      lt_framen   : in  std_logic;
      lt_ackn     : IN  std_logic;
      lt_dxfrn    : IN  std_logic;
      lt_tsr      : IN  std_logic_vector(11 downto 0);
      sram32_a    : OUT std_logic_vector(23 DOWNTO 0);
      sram32_we   : OUT std_logic;
      sram32_oe   : OUT std_logic;
      sram64_a    : OUT std_logic_vector(21 DOWNTO 0);
      sram64_oe   : OUT std_logic_vector(1 DOWNTO 0);
      sram64_we   : OUT std_logic_vector(1 DOWNTO 0);
      sram64_cs   : OUT std_logic;
      pci_out     : OUT std_logic
    );
  END COMPONENT;

  COMPONENT detect_64bit
    PORT (
      clk    : IN  std_logic;
      rstn   : IN  STD_LOGIC;
      req64n : IN  STD_LOGIC;
      det64  : OUT STD_LOGIC
     );
  END COMPONENT;

    COMPONENT IRQ_signaler
      GENERIC(
          irq_status_in_ptr      : integer RANGE 0 TO 255;
          pulse_length_in_ptr    : integer RANGE 0 TO 255
      );
      PORT (
		clock   : in  std_logic; -- PCI clock.
		resetn   : in  std_logic; -- PCI core reset signal.
		irq_generate : in std_logic; -- Signalizes that an interrupt is to be generated
		irq_status_read : in std_logic; -- If asserted then a target read has succeeded
		irq_line : out std_logic; -- This must connect to the IRQ line of the PCI core.

		dma_buffer_full : in std_logic;
		end_of_event    : in std_logic;

		-- config IF
		conf_addr    : IN  std_logic_vector(7 DOWNTO 0);
  		conf_in      : IN  std_logic_vector(31 DOWNTO 0);
		conf_out     : OUT std_logic_vector(31 DOWNTO 0);
		conf_we      : IN  std_logic
		);

    END COMPONENT;

  -------------------------------------------------------------------------------
  -- Type declaration
  -------------------------------------------------------------------------------

  TYPE input_state_type IS (fifo, store1, store2, latency_store1); -- OT 17.02.05

  -------------------------------------------------------------------------------
  -- Signal declaration
  -------------------------------------------------------------------------------
  SIGNAL resetn        : std_logic;             -- global reset
  SIGNAL rstn		       : std_logic; 				-- OT 19.04.05
  SIGNAL s_rst         : std_logic;             -- synchronous rese

  SIGNAL lt_rdyn_target    : std_logic;
  SIGNAL lt_rdyn_pci2mem   : std_logic;
  SIGNAL lt_abortn_target  : std_logic;
  SIGNAL lt_abortn_pci2mem : std_logic;
  SIGNAL lt_discn_target   : std_logic;
  SIGNAL lt_discn_pci2mem  : std_logic;

  SIGNAL pci_out_target  : std_logic;
  SIGNAL pci_out_master  : std_logic;
  SIGNAL pci_out_pci2mem : std_logic;

  SIGNAL conf_addr     : std_logic_vector(7 DOWNTO 0);
  SIGNAL conf_wr       : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_we       : std_logic;
  SIGNAL conf_rd       : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_rd_csr   : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_rd_dma   : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_rd_pg    : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_rd_ddl   : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_rd_evmon : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_rd_master: std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_rd_ram   : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_rd_irq: std_logic_vector(31 DOWNTO 0); -- OT 18.02.05

  SIGNAL mstr_done     : std_logic;
  SIGNAL start_mstr    : std_logic;
  SIGNAL last_block    : std_logic;
  SIGNAL burst_no      : std_logic_vector(31 DOWNTO 0);
  SIGNAL dma_addr      : std_logic_vector(31 DOWNTO 0);
  SIGNAL rpt_addr      : std_logic_vector(31 DOWNTO 0);
  SIGNAL rd_ptr        : std_logic_vector(31 DOWNTO 0);

  SIGNAL ddl_data  : std_logic_vector(31 DOWNTO 0);
  SIGNAL ddl_valid : std_logic;
  SIGNAL ddl_eoe   : std_logic;
  SIGNAL pg_data   : std_logic_vector(31 DOWNTO 0);
  SIGNAL pg_valid  : std_logic;
  SIGNAL pg_eoe    : std_logic;
  SIGNAL ram_data  : std_logic_vector(31 DOWNTO 0);
  SIGNAL ram_valid : std_logic;
  SIGNAL ram_eoe   : std_logic;
    
  SIGNAL busy      : std_logic;
  SIGNAL busy_en   : std_logic;
  SIGNAL pg_en     : std_logic;
  SIGNAL ram_en    : std_logic;
  SIGNAL ddl_en    : std_logic;
  SIGNAL dma_en    : std_logic;
  SIGNAL irq_en    : std_logic;  -- OT 18.02.05
  SIGNAL irq_generate : std_logic;  -- OT 18.02.05
  signal irq_line : std_logic; -- OT 18.02.05: This must connect to the IRQ line of the PCI core.
  SIGNAL pci_64bit : std_logic;

  SIGNAL fifo_out       : std_logic_vector(64 DOWNTO 0);
  SIGNAL fifo_in_data   : std_logic_vector(31 DOWNTO 0);
  SIGNAL fifo_in_eoe    : std_logic;
  SIGNAL fifo_in_valid  : std_logic;
  SIGNAL fifo_rd        : std_logic;
  SIGNAL buffer_wr        : std_logic;  -- OT 22.02.05
  SIGNAL fifo_rd_mstr   : std_logic;
  SIGNAL fifo_rd_pack   : std_logic;
  SIGNAL fifo_full      : std_logic;
  SIGNAL fifo_empty     : std_logic;
  SIGNAL fifo_clear     : std_logic;
  SIGNAL fifo_usedw     : std_logic_vector(7 DOWNTO 0);
  SIGNAL host_full      : std_logic;
  SIGNAL packetizer_in  : std_logic_vector(64 DOWNTO 0);
  SIGNAL packetizer_out : std_logic_vector(63 DOWNTO 0);
  SIGNAL dma_busy       : std_logic;

  SIGNAL eoe_count    : std_logic_vector(7 DOWNTO 0);
  SIGNAL event_count  : std_logic_vector(15 DOWNTO 0);
  SIGNAL event_length : std_logic_vector(31 DOWNTO 0);
  SIGNAL sel          : std_logic_vector(2 DOWNTO 0);

  SIGNAL data_store1    : std_logic_vector(64 DOWNTO 0);
  SIGNAL data_store2    : std_logic_vector(64 DOWNTO 0);
  SIGNAL burst_continue : std_logic;

  SIGNAL sram32_a_int      : std_logic_vector(23 DOWNTO 0);
  SIGNAL sram32_a_reg      : std_logic_vector(23 DOWNTO 0);
  SIGNAL sram32_d_inreg    : std_logic_vector(31 DOWNTO 0);
  SIGNAL sram32_d_outreg   : std_logic_vector(31 DOWNTO 0);
  SIGNAL sram32_we_int     : std_logic;
  SIGNAL sram32_we_reg     : std_logic;
  SIGNAL sram32_a_pci2mem  : std_logic_vector(23 DOWNTO 0);
  SIGNAL sram32_oe_pci2mem : std_logic;

  SIGNAL sram32_a_memif    : std_logic_vector(23 DOWNTO 0);
  SIGNAL sram32_oe_memif   : std_logic;

  SIGNAL sram64_a_int      : std_logic_vector(21 DOWNTO 0);
  SIGNAL sram64_a_reg      : std_logic_vector(21 DOWNTO 0);
  SIGNAL sram64_d_inreg    : std_logic_vector(63 DOWNTO 0);
  SIGNAL sram64_d_outreg   : std_logic_vector(63 DOWNTO 0);
  SIGNAL sram64_we_int     : std_logic_vector(1 DOWNTO 0);
  SIGNAL sram64_we_reg     : std_logic_vector(1 DOWNTO 0);
  SIGNAL sram64_a_pci2mem  : std_logic_vector(21 DOWNTO 0);
  SIGNAL sram64_oe_pci2mem : std_logic_vector(1 DOWNTO 0);
  SIGNAL sram64_cs_pci2mem : std_logic;

  SIGNAL mem_if_active     : std_logic;

  SIGNAL packetizer_input_state : input_state_type; -- OT 17.02.05

  signal dma_buffer_full : std_logic;
  signal end_of_event    : std_logic;

  -- New signals inferred from Xilinx implementation

  signal s_term_reg  : std_logic;
  signal s_ready_reg  : std_logic;
  signal m_ready_reg  : std_logic;


  signal oe_io_reg    : std_logic;
  -- These signals will be removed later

signal    l_cbeni         : std_logic_vector(7 downto 0);
signal    l_dato          : std_logic_vector(63 downto 0);
signal    l_adro          : std_logic_vector(63 downto 0);
signal    l_beno          : std_logic_vector(7 downto 0);
signal    l_cmdo          : std_logic_vector(3 downto 0);

signal    l_ldat_ackn     : std_logic;
signal    l_hdat_ackn     : std_logic;


signal    lt_abortn       : std_logic;
signal    lt_discn        : std_logic;
signal    lt_rdyn         : std_logic;
signal    lt_framen       : std_logic;
signal    lt_ackn         : std_logic;
signal    lt_dxfrn        : std_logic;
signal    lt_tsr          : std_logic_vector(11 downto 0);

signal    stat_reg     			: std_logic_vector(5 DOWNTO 0);

BEGIN  -- rtl

  resetn   <= rstn AND NOT (s_rst);
  INTR_N    <= irq_line;

  -- LEDs
  hostbuf_full <= host_full;
  dmabuf_full  <= fifo_full;
  dma_enabled  <= dma_en;
  dma_active   <= dma_busy;
  link_down    <= riLD_N;


  ---------------------------------------------------------------------------
  -- Interface to SRAM_32
  ---------------------------------------------------------------------------
  PROCESS (clk, rstn)
  BEGIN
    IF (rstn='0') THEN
      sram32_a_reg    <= (OTHERS => '0');
      sram32_d_inreg  <= (OTHERS => '0');
      sram32_d_outreg <= (OTHERS => '0');
      sram32_we_reg   <= '1';
    ELSIF rising_edge(clk) THEN
      sram32_a_reg    <= sram32_a_int;
      sram32_d_inreg  <= sram32_d;
      sram32_d_outreg <= l_dato(31 DOWNTO 0);
      sram32_we_reg   <= sram32_we_int;
    END IF;
  END PROCESS;
  -- multiplexed control signals
  sram32_a_int <= sram32_a_memif  WHEN (mem_if_active='1') ELSE sram32_a_pci2mem;
  sram32_oe    <= sram32_oe_memif WHEN (mem_if_active='1') ELSE sram32_oe_pci2mem;
  ---- control signals
  sram32_we <= sram32_we_reg;
  sram32_be <= "0000";
  ---- address/data
  sram32_a  <= sram32_a_reg;
  sram32_d  <= sram32_d_outreg WHEN (sram32_we_reg='0') ELSE (OTHERS => 'Z');



  ---------------------------------------------------------------------------
  -- Interface to SRAM_64
  ---------------------------------------------------------------------------
  PROCESS (clk, rstn)
  BEGIN
    IF (rstn='0') THEN
      sram64_a_reg    <= (OTHERS => '0');
      sram64_d_inreg  <= (OTHERS => '0');
      sram64_d_outreg <= (OTHERS => '0');
      sram64_we_reg   <= "11";
    ELSIF rising_edge(clk) THEN
      sram64_a_reg    <= sram64_a_int;
      sram64_d_inreg  <= sram64_d;
      sram64_d_outreg <= l_dato(63 DOWNTO 0);
      sram64_we_reg   <= sram64_we_int;
    END IF;
  END PROCESS;
  ---- multiplexor
  sram64_a_int <= sram64_a_pci2mem;

  ---- control signals
  sram64_oe <= sram64_oe_pci2mem;
  sram64_we <= sram64_we_reg;
  sram64_be <= "00000000";
  sram64_cs <= sram64_cs_pci2mem;
  ---- address/data
  sram64_a  <= sram64_a_reg;
  sram64_d  <= sram64_d_outreg WHEN (sram64_we_reg(0)='0' OR sram64_we_reg(1)='0') ELSE (OTHERS => 'Z');

  -----------------------------------------------------------------------------
  data_storage_for_disconnects:
  -----------------------------------------------------------------------------
  PROCESS (CLK_REG, rstn)
    BEGIN
      IF (rstn = '0') THEN
        data_store1 <= (OTHERS => '0');
        data_store2 <= (OTHERS => '0');
      ELSIF (CLK_REG'event and CLK_REG = '1') THEN
        IF (buffer_wr='1') THEN -- OT 22.02.05
          data_store1(64) <= '0';
          data_store1(63 downto 0) <= packetizer_out(63 downto 0); -- OT 22.02.05
          data_store2 <= data_store1;
        ELSE
          data_store1 <= data_store1;
          data_store2 <= data_store2;
        END IF;
      END IF;
    END PROCESS;

  u00: csr_block
    GENERIC MAP (
      global_csr => ad_csr_global,
      global_rst => ad_csr_reset,
      fclear_cmd => ad_csr_fclear
    )
    PORT MAP (
      clk          => clk,
      rstn         => rstn,
      s_rst        => s_rst,
      busy_en      => busy_en,
      pg_en        => pg_en,
      ram_en       => ram_en,
      ddl_en       => ddl_en,
      dma_en       => dma_en,
      irq_en       => irq_en,
      fifo_clear   => fifo_clear,
      dma_busy     => dma_busy,
      fifo_usedw   => fifo_usedw,
      fifo_full    => fifo_full,
      fifo_empty   => fifo_empty,
      hostbuf_full => host_full,
      pci_64bit    => pci_64bit,
      conf_addr    => conf_addr,
      conf_in      => conf_wr,
      conf_out     => conf_rd_csr,
      conf_we      => conf_we
    );

  pci2mem_1: pci2mem
    GENERIC MAP (
      bar_mem32 => 1,
      bar_mem64 => 2)
    PORT MAP (
      clk         => clk,
      rstn        => resetn,
      l_adro      => l_adro,
      l_cmdo      => l_cmdo,
      l_ldat_ackn => l_ldat_ackn,
      l_hdat_ackn => l_hdat_ackn,
      lt_abortn   => lt_abortn_pci2mem,
      lt_discn    => lt_discn_pci2mem,
      lt_rdyn     => lt_rdyn_pci2mem,
      lt_framen   => lt_framen,
      lt_ackn     => lt_ackn,
      lt_dxfrn    => lt_dxfrn,
      lt_tsr      => lt_tsr,
      sram32_a    => sram32_a_pci2mem,
      sram32_we   => sram32_we_int,
      sram32_oe   => sram32_oe_pci2mem,
      sram64_a    => sram64_a_pci2mem,
      sram64_oe   => sram64_oe_pci2mem,
      sram64_we   => sram64_we_int,
      sram64_cs   => sram64_cs_pci2mem,
      pci_out     => pci_out_pci2mem
    );

  u01: target_ctrl
    PORT MAP (
      clk           => CLK_REG, -- Using regional clock
      rstn	    => resetn,
      ADDR          => ADDR,
      ADIO          => ADIO,
      BASE_HIT      => BASE_HIT,
      S_ABORT       => S_ABORT,
      S_WRDN        => S_WRDN,
      IDLE          => IDLE,
      S_DATA_VLD    => S_DATA_VLD,
      pci_out       => pci_out_target,
      conf_wr       => conf_wr,
      conf_rd       => conf_rd,
      conf_addr     => conf_addr,
      conf_we       => conf_we
    );

  u0: master_ctrl
    GENERIC MAP (
      rd_ptr_addr_reg_addr  => ad_rd_ptr_addr_reg,
      rd_ptr_reg_addr       => ad_rd_ptr_reg,
      dma_error_reg_addr    => ad_dma_error_reg,
      report_error_reg_addr => ad_report_error_reg,
      read_error_reg_addr   => ad_read_error_reg
    )
    PORT MAP (
      clk           => CLK_REG, -- Using regional clock
      rstn          => resetn,
    ADIO            => ADIO,
    CSR             => CSR,
    M_DATA          => M_DATA,
    M_DATA_VLD      => M_DATA_VLD,
    M_ADDR_N        => M_ADDR_N,
    M_CBE           => M_CBE,
    M_WRDN          => M_WRDN,
    M_SRC_EN        => M_SRC_EN,
    M_FAIL64        => M_FAIL64,
    TIME_OUT        => TIME_OUT,
    COMPLETE        => COMPLETE,
    REQUEST         => REQUEST,
    REQUEST64       => REQUEST64,
    REQUESTHOLD     => REQUESTHOLD,
      stat_reg      => stat_reg,
      dma_addr      => dma_addr,
      rpt_addr      => rpt_addr,
      rd_ptr        => rd_ptr,
      start_mstr    => start_mstr,
      mstr_done     => mstr_done,
      last_block    => last_block,
      hostbuf_full  => host_full,
      burst_count   => burst_no,
      event_length  => event_length(18 DOWNTO 0),
      pci_out       => pci_out_master,
      oe_io_reg       => oe_io_reg,
      fifo_rd       => fifo_rd_mstr,
      burst_cont    => burst_continue,
      irq_en        => irq_en,
      irq_generate    => irq_generate,
      dma_buffer_full => dma_buffer_full,
      end_of_event  => end_of_event,
      conf_in       => conf_wr,
      conf_out      => conf_rd_master,
      conf_addr     => conf_addr,
      conf_we       => conf_we
    );

  u1: dma_engine
    GENERIC MAP (
      dma_wrptr      => ad_dma_writeptr,
      dma_buflen     => ad_dma_bufflen,
      dma_rptptr     => ad_dma_readptr,
      dma_rptptr_len => ad_dma_readptrlen,
      dma_status     => ad_dma_status_reg
    )
    PORT MAP (
      clk          => clk,
      rstn         => resetn,
      mstr_done    => mstr_done,
      start_mstr   => start_mstr,
      dma_addr     => dma_addr,
      rpt_addr     => rpt_addr,
      rd_ptr       => rd_ptr,
      dma_en       => dma_en,
      dma_busy     => dma_busy,
      fifo_usedw   => fifo_usedw,
      fifo_full    => fifo_full,
      burst_no     => burst_no,
      last_block   => last_block,
      hostbuf_full => host_full,
      conf_addr    => conf_addr,
      conf_in      => conf_wr,
      conf_out     => conf_rd_dma,
      conf_we      => conf_we
    );

  u2: pattern_generator
    GENERIC MAP (
      patgen_evlen_reg => ad_patgen_evlen_reg,
      patgen_csr_reg   => ad_patgen_csr_reg,
      patgen_start_reg => ad_patgen_start_reg
    )
    PORT MAP (
      clk       => clk,
      rstn      => resetn,
      data      => pg_data,
      d_valid   => pg_valid,
      eoe_flag  => pg_eoe,
      busy_n    => busy,
      conf_addr => conf_addr,
      conf_in   => conf_wr,
      conf_out  => conf_rd_pg,
      conf_we   => conf_we
    );

  u3: dma_fifo
    PORT MAP (
      clk        => CLK_REG,
      clk_global => CLK,
      rstn       => resetn,
      data       => fifo_in_data,
      d_valid    => fifo_in_valid,
      eoe_flag   => fifo_in_eoe,
      fifo_rd    => fifo_rd,
      fifo_clear => fifo_clear,
      fifo_out   => fifo_out,
      full       => fifo_full,
      empty      => fifo_empty,
      usedw      => fifo_usedw,
      event_length => event_length
    );

  u4: ddl_if
    GENERIC MAP (
      commd_reg_addr => ad_commd_reg,
      ctstw_reg_addr => ad_ctstw_reg,
      festw_reg_addr => ad_festw_reg,
      dtstw_reg_addr => ad_dtstw_reg,
      ifstw_reg_addr => ad_ifstw_reg
    )
    PORT MAP (
      clk          => clk,
      rstn         => resetn,
      roCLK        => roCLK,
      roRST_N      => roRST_N,
      riLD_N       => riLD_N,
      riLF_N       => riLF_N,
      roBSY_N      => roBSY_N,
      roD          => roD,
      roTEN_N      => roTEN_N,
      roCMD_N      => roCMD_N,
      riD          => riD,
      riTEN_N      => riTEN_N,
      riSTS_N      => riSTS_N,
      data         => ddl_data,
      d_valid      => ddl_valid,
      eoe_flag     => ddl_eoe,
      busy_n       => busy,
      conf_addr    => conf_addr,
      conf_in      => conf_wr,
      conf_out     => conf_rd_ddl,
      conf_we      => conf_we
    );

    mem_if_1: mem_if
      GENERIC MAP (
        memif_addr    => ad_memif_addr,
        memif_length  => ad_memif_length,
        memif_csr     => ad_memif_csr,
        memif_trigger => ad_memif_trigger)
      PORT MAP (
        clk       => clk,
        rstn      => resetn,
        addr      => sram32_a_memif,
        d         => sram32_d_inreg,
        oe        => sram32_oe_memif,
        data      => ram_data,
        d_valid   => ram_valid,
        eoe_flag  => ram_eoe,
        busy_n    => busy,
        active    => mem_if_active,
        conf_addr => conf_addr,
        conf_in   => conf_wr,
        conf_out  => conf_rd_ram,
        conf_we   => conf_we
      );

  u5: packetizer
    PORT MAP (
      clk        => CLK_REG,
      rstn       => resetn,
      rdreq_in   => fifo_rd_mstr,
      rdreq_out  => fifo_rd_pack,
      burst_continue  => burst_continue,
      mstr_done  => mstr_done,
      last_block => last_block,
      data_in    => packetizer_in,
      data_out   => packetizer_out
    );

    u6: IRQ_signaler
    GENERIC MAP (
      irq_status_in_ptr => ad_irq_status,
      pulse_length_in_ptr => ad_pulse_length
    )
    PORT MAP (
      clock    => clk,
      resetn   => resetn,
      irq_generate => irq_generate,
      irq_status_read => S_DATA_VLD,
      irq_line  => irq_line,
      dma_buffer_full => dma_buffer_full,
      end_of_event => end_of_event,
      conf_addr  => conf_addr,
      conf_in  => conf_wr,
      conf_out  => conf_rd_irq,
      conf_we  => conf_we
    );

  event_monitor_1: event_monitor
    GENERIC MAP (
      evmon_counter => evmon_counter_addr)
    PORT MAP (
      clk          => clk,
      rstn         => resetn,
      data         => fifo_out,
      eoe_count    => eoe_count,
      event_count  => event_count,
      conf_addr    => conf_addr,
      conf_in      => conf_wr,
      conf_out     => conf_rd_evmon,
      conf_we      => conf_we
    );

  rstn <= not RST; -- OT 19.04.05


  -------------------------------------------------------------------------------
  -- Data path
  -------------------------------------------------------------------------------

  -- config IF
  conf_rd <= conf_rd_csr OR conf_rd_dma OR conf_rd_pg OR conf_rd_ddl OR conf_rd_master OR
             conf_rd_evmon OR conf_rd_ram OR conf_rd_irq; -- OT 18.02.05: Added conf_rd_irq

  -- PCI
  lt_rdyn   <= lt_rdyn_target   AND lt_rdyn_pci2mem;
  lt_discn  <= lt_discn_target  AND lt_discn_pci2mem;
  lt_abortn <= lt_abortn_target AND lt_abortn_pci2mem;

  -- l_adi select
  PROCESS (pci_out_master, pci_out_target, pci_out_pci2mem, oe_io_reg,
           sram32_d_inreg, sram64_d_inreg, sram64_oe_pci2mem, packetizer_out)
  BEGIN
    IF (pci_out_pci2mem='1') then
      IF (sram64_oe_pci2mem = "00") THEN
        ADIO <= sram64_d_inreg;
      ELSE
        ADIO <= sram32_d_inreg & sram32_d_inreg;
      END IF;
    ELSIF (pci_out_master='1') then
      ADIO <= (others=>'Z'); -- master_ctrl driving ADIO by itself
    ELSIF (pci_out_target='1') then
      ADIO <= (others=>'Z'); -- target_ctrl driving ADIO by itself
    ELSE
      IF (oe_io_reg='1') THEN
            IF (packetizer_input_state=fifo) THEN -- OT 23.02.05
                  	ADIO <= packetizer_out;
            ELSE
                        ADIO <= packetizer_in(63 downto 0);
            END IF;
       ELSE
            ADIO <= (others=>'Z');
       END IF;
    END IF;
  END PROCESS;


  -- DMA-FIFO input
  fifo_in_data  <= pg_data  WHEN (pg_en  = '1') ELSE
                   ram_data WHEN (ram_en = '1') ELSE
                   ddl_data WHEN (ddl_en = '1') ELSE
                   (OTHERS => '0');
  fifo_in_eoe   <= pg_eoe   WHEN (pg_en  = '1') ELSE
                   ram_eoe  WHEN (ram_en = '1') ELSE
                   ddl_eoe  WHEN (ddl_en = '1') ELSE
                   '0';
  fifo_in_valid <= pg_valid  WHEN (pg_en  = '1') ELSE
                   ram_valid WHEN (ram_en = '1') ELSE
                   ddl_valid WHEN (ddl_en = '1') ELSE
                   '0';
--  fifo_rd       <= (fifo_rd_pack AND (NOT burst_continue)) WHEN (packetizer_input_state=fifo) ELSE '0'; -- OT 17.02.05: Added conditions
--  buffer_wr       <= (fifo_rd_mstr AND (NOT burst_continue)) WHEN (packetizer_input_state=fifo) ELSE '0';

  fifo_rd       <= (fifo_rd_pack) WHEN (packetizer_input_state=fifo) ELSE '0'; -- OT 17.02.05: Added conditions
  buffer_wr       <= (fifo_rd_mstr) WHEN (packetizer_input_state=fifo) ELSE '0';


  -- data store
--  packetizer_in <= data_store2 WHEN (burst_continue='1') ELSE fifo_out;

	packetizer_in <= data_store2 WHEN (packetizer_input_state=store2) ELSE
			 data_store1 WHEN (packetizer_input_state=store1 OR packetizer_input_state=latency_store1) ELSE fifo_out;

  -----------------------------------------------------------------------------
  state_machine_controlling_packetizer_input:
  -----------------------------------------------------------------------------
  PROCESS (clk_reg, rstn)
    BEGIN
      IF (rstn = '0') THEN
	packetizer_input_state <= fifo;
      ELSIF (clk_reg'event and clk_reg = '1') THEN
	CASE packetizer_input_state IS

          WHEN fifo =>
            IF (burst_continue = '1') THEN -- If the former transaction ended with a disconnect, we have to resend the words stored in the buffers
		IF (TIME_OUT='1') THEN
		  packetizer_input_state <= latency_store1; -- Only resending data_store1 if the disconnect was due to a latency timer expiration
		ELSE
      		  packetizer_input_state <= store2;
      		END IF;
            ELSE
              packetizer_input_state <= fifo;
            END IF;

          WHEN store2 =>
            IF ((fifo_rd_pack = '1') OR (fifo_rd_mstr='1')) THEN
              packetizer_input_state <= store1;
            ELSE
              packetizer_input_state <= store2;
            END IF;

          WHEN store1 =>
            IF (M_DATA='1') THEN -- If the word stored in data_store1 was successfully sent, we return to fifo state
              packetizer_input_state <= fifo;
            ELSE
              packetizer_input_state <= store1;
            END IF;

          WHEN latency_store1 =>
            IF ((fifo_rd_pack = '1') OR (fifo_rd_mstr='1')) THEN -- If the word stored in data_store1 was successfully sent, we return to fifo state
              packetizer_input_state <= fifo;
            ELSE
              packetizer_input_state <= latency_store1;
            END IF;

          WHEN others =>
          packetizer_input_state <= fifo;

        END CASE;

      END IF;
    END PROCESS;





  busy <= '0' WHEN ((fifo_usedw = "11111100" OR
                     fifo_usedw = "11111101" OR
                     fifo_usedw = "11111110" OR
                     fifo_usedw = "11111111" OR
                     fifo_full ='1') AND busy_en = '1') ELSE '1';



  -----------------------------------------------------------------------
  -- This section contains unused signals.
  -----------------------------------------------------------------------

  C_TERM <= '1';
  C_READY <= '1';
  KEEPOUT <= '0';
  CFG_SELF <= '0';
  SUB_DATA <= "00000000000000000000000000000000";

  -- The following registered signals must NOT be assigned to static values:

  DONT_OPTIMIZE_PLEASE:
  process(RST, CLK)
  begin
    if RST = '1' then
      s_term_reg <= '1';
      s_ready_reg <= '0';
      m_ready_reg <= '0';
    elsif (CLK'event and CLK = '1') then
      s_term_reg <= '1'; -- Always single phase on target transactions
      s_ready_reg <= '1';
      m_ready_reg <= '1';
    end if;
  end process;

  S_TERM <= s_term_reg;
  S_READY <= s_ready_reg;
  M_READY <= m_ready_reg;
  SLOT64 <= '1';

END rtl;
