--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : master_ctrl.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de> / Artur Szostak / Olav Torheim <torheim@ift.uib.no>
-- Company    :
-- Platform   :
-------------------------------------------------------------------------------
-- Description:
--
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author      Description
--
-- 07.07.2004  1.0      tilsner     created
--
-- 09.03.2005  1.1      A. Szostak  conf_addr is no longer compared to
--                                  hardcoded values. Fixed some typos.
--
-- 24.06.2005  1.2      O. Torheim  Replacing the Altera MegaCore interface with the Xilinx LogiCore interface
--                                  Adding error handling (latency timer expiration and retries on the PCI bus)
--                                  Adding logic for interrupt generation
--=============================================================================

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;

ENTITY master_ctrl IS

  GENERIC (
    rd_ptr_addr_reg_addr  : integer RANGE 0 TO 255 := 0;
    rd_ptr_reg_addr       : integer RANGE 0 TO 255 := 4;
    dma_error_reg_addr    : integer RANGE 0 TO 255 := 8;
    report_error_reg_addr : integer RANGE 0 TO 255 := 12;
    read_error_reg_addr   : integer RANGE 0 TO 255 := 16
  );

  PORT (
    --system
    clk          : in std_logic;
    rstn         : in std_logic;
    ADIO            : inout std_logic_vector( 63 downto 0);
    CSR             : in    std_logic_vector( 39 downto 0);
    M_DATA          : in    std_logic;
    M_DATA_VLD      : in    std_logic;
    M_ADDR_N        : in    std_logic;
    M_CBE           : out   std_logic_vector(  7 downto 0);
    M_WRDN          : out   std_logic;
    M_SRC_EN        : in    std_logic;
    M_FAIL64        : in    std_logic;
    TIME_OUT        : in    std_logic;
    COMPLETE        : out   std_logic;
    REQUEST         : out   std_logic;
    REQUEST64       : out   std_logic;
    REQUESTHOLD     : out   std_logic;

    --master control

    stat_reg     : IN std_logic_vector(5 DOWNTO 0);

    --dma engine interface
    dma_addr     : IN  std_logic_vector(31 downto 0);
    rpt_addr     : IN  std_logic_vector(31 DOWNTO 0);
    rd_ptr       : OUT std_logic_vector(31 DOWNTO 0);
    start_mstr   : IN  std_logic;
    mstr_done    : OUT std_logic;
    last_block   : IN  std_logic;
    hostbuf_full : IN  std_logic;
    burst_count  : IN std_logic_vector(31 DOWNTO 0);
    event_length : IN std_logic_vector(18 DOWNTO 0);

    pci_out      : OUT std_logic;
    oe_io_reg    : INOUT std_logic;
    fifo_rd      : OUT std_logic;

    burst_cont   : OUT std_logic;

    --irq interface
    irq_en       : IN std_logic; -- OT 18.02.05
    irq_generate : OUT std_logic; -- OT 18.02.05

    -- Reason bits for raising an IRQ.
    dma_buffer_full : out std_logic;
    end_of_event    : out std_logic;

    -- config IF
    conf_addr    : IN  std_logic_vector(7 DOWNTO 0);
    conf_in      : IN  std_logic_vector(31 DOWNTO 0);
    conf_out     : OUT std_logic_vector(31 DOWNTO 0);
    conf_we      : IN  std_logic
  );

END master_ctrl;




ARCHITECTURE behavior OF master_ctrl is

  TYPE state_type IS (idle, req_bus, mem_wr_addr_phase, mem_wr_data_phase,
                      report_req, report_addr, report_data_rdy, report_data_out, report_retry, -- OT 17.02.05: Added report_retry
                      irq_req, irq_wait, read_req, rd_addr_phase, rd_data_phase, rd_retry, pci_disc); -- OT 18.02.05: Added irq_req and irq_wait
  SIGNAL state : state_type;

  TYPE last_state_type IS (idle, mw_wait, last);
  SIGNAL last_state : last_state_type;

  TYPE disc_state_type IS (disc_idle, disc_wod, disc_wd, disc_wait1, disc_latency, disc_retry_burst); -- OT 17.02.05: Added disc_latency and disc_retry_burst
  SIGNAL disc_state : disc_state_type;

  SIGNAL we                    : std_logic_vector(1 DOWNTO 0);
  SIGNAL disengage_mstr        : std_logic;
  SIGNAL qword_cnt             : std_logic_vector(31 DOWNTO 0);
  SIGNAL data_phase_fall       : std_logic;
  SIGNAL data_phase            : std_logic;
  SIGNAL data_phase_reg        : std_logic;
  SIGNAL req32n                : std_logic;
  SIGNAL req64n                : std_logic;

  SIGNAL ADIO_i       : std_logic_vector(63 DOWNTO 0);
  SIGNAL ADIO_reg     : std_logic_vector(63 DOWNTO 0);
  SIGNAL M_CBE_i     : std_logic_vector(7 DOWNTO 0);
  SIGNAL M_CBE_reg   : std_logic_vector(7 DOWNTO 0);

  SIGNAL REQUEST_i : std_logic;
  SIGNAL REQUEST64_i : std_logic;
  SIGNAL REQUEST_reg : std_logic;
  SIGNAL REQUEST64_reg : std_logic;

  SIGNAL COMPLETE_i    : std_logic;
  SIGNAL COMPLETE_reg  : std_logic;

  SIGNAL rd_ptr_reg  : std_logic_vector(31 DOWNTO 0);
  SIGNAL rd_ptr_addr : std_logic_vector(31 DOWNTO 0);
  SIGNAL load_rd_ptr : std_logic;

  SIGNAL retry_count   : std_logic_vector(4 DOWNTO 0);
  SIGNAL timeout_count : std_logic_vector(4 DOWNTO 0);
  SIGNAL delay_count   : std_logic_vector(7 DOWNTO 0);

  SIGNAL pci_addr_cnt   : std_logic_vector(31 DOWNTO 0);
  SIGNAL next_pci_addr  : std_logic_vector(31 DOWNTO 0);
  SIGNAL burst_continue : std_logic;

  SIGNAL dma_error    : std_logic_vector(31 DOWNTO 0);
  SIGNAL report_error : std_logic_vector(31 DOWNTO 0);
  SIGNAL read_error   : std_logic_vector(31 DOWNTO 0);

  SIGNAL buf_full_irq : std_logic;
  SIGNAL eoe_irq      : std_logic;

  SIGNAL fatal_reg      : std_logic; -- OT 20. april 2005
  SIGNAL retry_reg      : std_logic; -- OT 20. april 2005
  SIGNAL disc_wd_reg    : std_logic; -- OT 20. april 2005
  SIGNAL disc_wod_reg    : std_logic; -- OT 20. april 2005

  BEGIN

    burst_cont  <= burst_continue;
    rd_ptr      <= rd_ptr_reg;

    irq_generate <= buf_full_irq or eoe_irq;
    dma_buffer_full <= buf_full_irq;
    end_of_event    <= eoe_irq;

    --disengage the master if
    --TIME_OUT -- lateny expired
    --retry_reg -- retry
    --disc_wod_reg -- disc_wod
    --disc_wd_reg -- disc_wd
    disengage_mstr <= TIME_OUT or retry_reg or disc_wd_reg or disc_wod_reg;

    COMPLETE_i <= '1' WHEN ((last_state = last) OR      -- signal for master burst write
                           (state=report_addr) OR  -- signal for master single cycle
                           (state=report_data_rdy) OR  -- signal for master single cycle
                           (state=report_data_out) OR  -- signal for master single cycle
                           (state=rd_data_phase) OR  -- signal for master single cycle
--			  (state=irq_wait AND lm_adr_ackn='1') OR -- OT 18.02.05: Added irq state conditions
                           (state=rd_addr_phase)) ELSE '0';

    pci_out <= '1' WHEN (state = mem_wr_addr_phase OR
                         state = report_addr       OR
                         state = report_data_rdy   OR
                         state = report_data_out   OR
                         state = rd_addr_phase)    ELSE '0';

      fifo_rd <= (M_SRC_EN) WHEN (NOT (state=rd_data_phase)) ELSE '0'; -- OT 19.02.05: We do not want to read from the FIFO when asking for the read pointer

    ---------------------------------------------------------------------------
    conf_if:
    ---------------------------------------------------------------------------
    we(0)    <= '1' WHEN (conf_we='1' AND conf_addr = rd_ptr_addr_reg_addr) ELSE '0';
    we(1)    <= '1' WHEN (conf_we='1' AND conf_addr = rd_ptr_reg_addr) ELSE '0';
    conf_out <= rd_ptr_addr  WHEN (conf_addr = rd_ptr_addr_reg_addr) ELSE
                rd_ptr_reg   WHEN (conf_addr = rd_ptr_reg_addr) ELSE
                dma_error    WHEN (conf_addr = dma_error_reg_addr) ELSE
                report_error WHEN (conf_addr = report_error_reg_addr) ELSE
                read_error   WHEN (conf_addr = read_error_reg_addr) ELSE
                (OTHERS => '0');

    ---------------------------------------------------------------------------
    pci_output_regs:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          ADIO_reg     <= (OTHERS => '0');
          M_CBE_reg   <= (OTHERS => '0');
          REQUEST_reg <= '0';
          REQUEST64_reg <= '0';
          COMPLETE_reg  <= '0';
        ELSIF (clk'event AND clk='1') THEN
          ADIO_reg     <= ADIO_i;
          M_CBE_reg   <= M_CBE_i;
          REQUEST64_reg <= REQUEST64_i;
          REQUEST_reg <= REQUEST_i;
          COMPLETE_reg  <= COMPLETE_i;
        END IF;
    END PROCESS;

    oe_io_reg <= '1' when ((state=mem_wr_addr_phase AND M_ADDR_N='0') OR (state=report_addr AND M_ADDR_N='0') OR (state=rd_addr_phase AND M_ADDR_N='0') OR (( (state=mem_wr_data_phase) OR (state=report_data_rdy) ) AND (M_DATA='1' OR (M_ADDR_N='0')) ) ) else '0';

    ADIO <= ADIO_reg when ((oe_io_reg = '1') AND (NOT (state=mem_wr_data_phase))) else (others=>'Z');

    M_CBE   <= M_CBE_reg when (M_ADDR_N = '0') else (others =>'0');
    REQUEST64 <= REQUEST64_reg;
    REQUEST <= REQUEST_reg;
    COMPLETE  <= COMPLETE_reg;

    ---------------------------------------------------------------------------
    error_register:
    ---------------------------------------------------------------------------      
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          dma_error    <= (OTHERS => '0');
          report_error <= (OTHERS => '0');
          read_error   <= (OTHERS => '0');
        ELSIF (clk'event AND clk='1') THEN
          CASE state IS
            WHEN req_bus | mem_wr_addr_phase | mem_wr_data_phase =>
              IF (TIME_OUT = '1') THEN
                dma_error(4 DOWNTO 0) <= dma_error(4 DOWNTO 0) + 1;
              END IF;
              IF (retry_reg = '1') THEN
                dma_error(9 DOWNTO 5) <= dma_error(9 DOWNTO 5) + 1;
              END IF;
              IF (disc_wod_reg = '1') THEN
                dma_error(14 DOWNTO 10) <= dma_error(14 DOWNTO 10) + 1;
              END IF;
              IF (disc_wd_reg = '1') THEN
                dma_error(19 DOWNTO 15) <= dma_error(19 DOWNTO 15) + 1;
              END IF;
              IF (stat_reg(2) = '1') THEN
                dma_error(24 DOWNTO 20) <= dma_error(24 DOWNTO 20) + 1;
              END IF;
              IF (stat_reg(3) = '1') THEN
                dma_error(29 DOWNTO 25) <= dma_error(29 DOWNTO 25) + 1;
              END IF;


            WHEN report_req | report_addr | report_data_rdy | report_data_out =>
              IF (TIME_OUT = '1') THEN
                report_error(4 DOWNTO 0) <= report_error(4 DOWNTO 0) + 1;
              END IF;
              IF (retry_reg = '1') THEN
                report_error(9 DOWNTO 5) <= report_error(9 DOWNTO 5) + 1;
              END IF;
              IF (disc_wod_reg = '1') THEN
                report_error(14 DOWNTO 10) <= report_error(14 DOWNTO 10) + 1;
              END IF;
              IF (disc_wd_reg = '1') THEN
                report_error(19 DOWNTO 15) <= report_error(19 DOWNTO 15) + 1;
              END IF;
              IF (stat_reg(2) = '1') THEN
                report_error(24 DOWNTO 20) <= report_error(24 DOWNTO 20) + 1;
              END IF;
              IF (stat_reg(3) = '1') THEN
                report_error(29 DOWNTO 25) <= report_error(29 DOWNTO 25) + 1;
              END IF;

            WHEN read_req | rd_addr_phase | rd_data_phase =>
              IF (TIME_OUT = '1') THEN
                read_error(4 DOWNTO 0) <= read_error(4 DOWNTO 0) + 1;
              END IF;
              IF (retry_reg = '1') THEN
                read_error(9 DOWNTO 5) <= read_error(9 DOWNTO 5) + 1;
              END IF;
              IF (disc_wod_reg = '1') THEN
                read_error(14 DOWNTO 10) <= read_error(14 DOWNTO 10) + 1;
              END IF;
              IF (disc_wd_reg = '1') THEN
                read_error(19 DOWNTO 15) <= read_error(19 DOWNTO 15) + 1;
              END IF;
              IF (stat_reg(2) = '1') THEN
                read_error(24 DOWNTO 20) <= read_error(24 DOWNTO 20) + 1;
              END IF;
              IF (stat_reg(3) = '1') THEN
                read_error(29 DOWNTO 25) <= read_error(29 DOWNTO 25) + 1;
              END IF;

            WHEN OTHERS =>

          END case;

        END IF;
      END PROCESS;


    -------------------------------------------------------------------------------
    rd_ptr_register:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          rd_ptr_reg  <= (OTHERS => '0');
          rd_ptr_addr <= (OTHERS => '0');
        ELSIF (clk'event AND clk='1') THEN
          IF (load_rd_ptr='1') THEN
            rd_ptr_reg <= ADIO(31 DOWNTO 0);
          END IF;
          IF (we(0)='1') THEN
            rd_ptr_addr <= conf_in;
          END IF;
          IF (we(1)='1') THEN
            rd_ptr_reg <= conf_in;
          END IF;
        END IF;
      END PROCESS;


    -------------------------------------------------------------------------------
    fsm_controller:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF rstn='0' THEN
        state <= idle;
      ELSIF clk'event AND clk='1' THEN
        CASE state IS

          WHEN idle =>  -- update rd_pointer by PCI master read cycle
            IF (hostbuf_full='1' AND delay_count="00000000" AND irq_en='0') THEN
              state <= read_req;
            ELSIF (hostbuf_full='1' AND irq_en='1') THEN -- OT 18.02.05: Added IRQ states
              state <= irq_req;
            ELSIF (start_mstr = '1') THEN
              state <= req_bus;
            ELSE
              state <= idle;
            END IF;

          -- memory write transaction
          WHEN req_bus =>
                        state <= mem_wr_addr_phase;

          -- memory write address phase
          WHEN mem_wr_addr_phase =>
            IF M_ADDR_N = '0' THEN
              state <= mem_wr_data_phase;
            ELSE
              state <= mem_wr_addr_phase;
            END IF;

          -- memory write data phase

          WHEN mem_wr_data_phase =>
            IF (disc_wod_reg='1' OR disc_wd_reg='1' OR (TIME_OUT='1' AND (NOT (qword_cnt=0)) AND (NOT (qword_cnt=1))) OR retry_reg='1') THEN -- OT 17.02.05: Added latency timer expiration and retry. OT 19.02.05: Skipping disconnect on special case of latency timer expiration at last planned data word
              state <= pci_disc;
--            ELSIF ((lm_tsr(9 downto 0)="000000000"  OR disengage_mstr='1') AND last_block='0') THEN
            ELSIF ((M_DATA='0'  OR disengage_mstr='1') AND last_block='0') THEN
              state <= idle;
--            ELSIF (( (lm_tsr(9 downto 4)="000000" AND lm_tsr(2 downto 0)="000")  OR disengage_mstr='1') AND lm_tsr(3)='0' AND last_block='1') THEN -- OT 24.02.05
            ELSIF ((M_DATA='0'  OR disengage_mstr='1') AND last_block='1') THEN
              state <= report_req;
            ELSE
              state <= mem_wr_data_phase;
            END IF;

          -- bus cycle for report word
          WHEN report_req =>
            state <= report_addr;

          WHEN report_addr =>
            IF M_ADDR_N = '0' THEN
              state <= report_data_rdy;
            ELSE
              state <= report_addr;
            END IF;

          WHEN report_data_rdy =>
            state <= report_data_out;

          WHEN report_data_out =>
	    IF (retry_reg='1') THEN -- OT 17.02.05: Added retry handling
              state <= report_retry;
--            ELSIF (lm_tsr(9 downto 0) = "000000000" OR disengage_mstr = '1') THEN
            ELSIF (M_DATA = '0' OR disengage_mstr = '1') THEN
              state <= idle;
            ELSE
              state <= report_data_out;
            END IF;

	  WHEN report_retry => -- OT 17.02.05: Retry handling for report word
            IF (retry_count(3)='1') THEN
              state <= idle;
	    ELSE
              state <= report_req;
            END IF;

          WHEN read_req =>
            state <= rd_addr_phase;

          WHEN rd_addr_phase =>
            IF (M_ADDR_N='0') THEN
              state <= rd_data_phase;
            ELSE
              state <= rd_addr_phase;
            END IF;

          WHEN rd_data_phase =>
            IF (retry_reg='1') THEN
              state <= rd_retry;
            ELSIF (M_DATA = '0' OR disengage_mstr = '1') THEN
              state <= idle;
            ELSE
              state <= rd_data_phase;
            END IF;
            
          WHEN rd_retry =>
            IF (retry_count(3)='1') THEN
              state <= idle;
            ELSIF (timeout_count(3)='1') THEN
              state <= read_req;
            ELSE
              state <= rd_retry;
            END IF;

          WHEN irq_req => -- OT 18.02.05: Added IRQ states
            state <= irq_wait;

          WHEN irq_wait =>
            IF (hostbuf_full='1') THEN
              state <= irq_wait;
            ELSE
              state <= idle;
            END IF;

-- 23.05.
--          WHEN pci_disc_wod =>
--            state <= req_bus;
          WHEN pci_disc =>
          	 if (retry_count(3)='0') then -- OT 18.02.05: We cannot reclaim the bus forever
            		state <= req_bus;
            else
            		state <= pci_disc;		-- OT 18.02.05: For now, we stay in state pci_disc forever. In later revisions, something needs to be done here.
              end if;

          WHEN others =>
            state <= idle;

        END CASE;
      END IF;
    END PROCESS;


    -------------------------------------------------------------------------------
    data_controller:
    -------------------------------------------------------------------------------
    PROCESS (state, dma_addr, rpt_addr, M_ADDR_N, M_DATA, M_DATA_VLD, disengage_mstr, last_block,
             burst_count, event_length, M_SRC_EN, rd_ptr_addr)
    BEGIN

      -- default values
      ADIO_i(63 downto 0)  <= (OTHERS => '0');
      M_CBE_i(7 downto 0) <= "11111111";
      REQUEST_i   <= '0';
      REQUEST64_i   <= '0';
      req32n        <= '1';
      req64n        <= '1';

      mstr_done   <= '0';
      load_rd_ptr <= '0';

      eoe_irq <= '0';

      CASE state IS

        WHEN idle =>
          --initialization
          ADIO_i(63 downto 0)  <= (OTHERS => '0');
          M_CBE_i(7 downto 0) <= "11111111";
          REQUEST64_i   <= '0';
          REQUEST_i   <= '0';
          req64n      <= '1';

          mstr_done   <= '0';
          buf_full_irq <= '0'; -- OT

        WHEN req_bus =>
          REQUEST64_i <= '1';
          req64n    <= '0';

          IF (burst_continue='1') THEN
            ADIO_i <= x"00000000" & next_pci_addr;
          ELSE
            ADIO_i <= x"00000000" & dma_addr;
          END IF;
          M_CBE_i(7 downto 0) <= "11110111";
          buf_full_irq <= '0'; -- OT


        WHEN mem_wr_addr_phase =>
          REQUEST64_i <= '0';
          req64n    <= '1';

          IF (burst_continue='1') THEN
            ADIO_i <= x"00000000" & next_pci_addr;
          ELSE
            ADIO_i <= x"00000000" & dma_addr;
          END IF;
--          IF (lm_adr_ackn = '0') THEN
--            M_CBE_i <= "00000000";
--          ELSE
            M_CBE_i <= "11110111";
--          END IF;
          buf_full_irq <= '0'; -- OT


        WHEN mem_wr_data_phase =>

--          IF (lm_tsr(9 downto 0) = "00000000" AND last_block = '0') THEN
          IF (M_DATA='0' AND last_block = '0') THEN
            mstr_done <= '1';
          ELSIF (TIME_OUT='1' AND (qword_cnt=0 OR qword_cnt=1) AND last_block='0') THEN -- OT 19.02.05: Added logic for handling special case of latency timer expiration at the last planned data phase of the transaction
            mstr_done <= '1';
          ELSE
            mstr_done <= '0';
          END IF;
          buf_full_irq <= '0'; -- OT


        WHEN report_req =>
          REQUEST_i <= '1';
          req32n      <= '0';

          ADIO_i(63 downto 0)  <= x"00000000" & rpt_addr;
          M_CBE_i(7 downto 0) <= "11110111";

          buf_full_irq <= '0'; -- OT


        WHEN report_addr =>
          REQUEST_i <= '0';
          req32n      <= '1';

--          ADIO_i(63 downto 0)  <= x"00000000" & rpt_addr;
--          M_CBE_i(7 downto 0) <= "11110111";
          IF (M_ADDR_N = '0') THEN
            ADIO_i(63 downto 0)  <= event_length & '0' & x"000" & event_length & '0' & x"000";
            M_CBE_i <= "00000000";
          ELSE
            ADIO_i(63 downto 0)  <= x"00000000" & rpt_addr;
            M_CBE_i <= "11110111";
          END IF;
          buf_full_irq <= '0'; -- OT


        WHEN report_data_rdy =>

--          l_adi_i(63 downto 0)  <= event_length & burst_count(12 DOWNTO 0) & event_length & burst_count(12 DOWNTO 0);
          ADIO_i(63 downto 0)  <= event_length & '0' & x"000" & event_length & '0' & x"000";
          M_CBE_i(7 downto 0) <= "00000000";
          buf_full_irq <= '0'; -- OT

        WHEN report_data_out =>

--          l_adi_i(63 downto 0)  <= event_length & burst_count(12 DOWNTO 0) & event_length & burst_count(12 DOWNTO 0);
          ADIO_i(63 downto 0)  <= event_length & '0' & x"000" & event_length & '0' & x"000";
          M_CBE_i(7 downto 0) <= "00000000";

          IF (M_DATA='0'  AND disengage_mstr = '0') THEN -- OT 17.02.05: Do NOT assert mstr_done if the transaction was abnormally terminated
            mstr_done <= '1';
	    eoe_irq <= '1';
          ELSE
            mstr_done <= '0';
	    eoe_irq <= '0';
          END IF;
          buf_full_irq <= '0'; -- OT


        WHEN read_req =>
          REQUEST_i <= '1';
          req32n      <= '0';

          ADIO_i   <= x"00000000" & rd_ptr_addr;
          M_CBE_i <= "11110110";
          buf_full_irq <= '0'; -- OT


        WHEN rd_addr_phase =>
          REQUEST_i <= '0';
          req32n      <= '1';

          ADIO_i(63 downto 0)  <= x"00000000" & rd_ptr_addr;
--          IF (lm_adr_ackn = '0') THEN
--            M_CBE_i <= "00000000";
--          ELSE
            M_CBE_i <= "11110110";
--          END IF;
          buf_full_irq <= '0'; -- OT


        WHEN rd_data_phase =>

          IF (M_DATA_VLD='1') THEN
            load_rd_ptr <= '1';
          END IF;
          buf_full_irq <= '0'; -- OT

        WHEN irq_req =>

          buf_full_irq <= '1'; -- OT

          ADIO_i(63 downto 0)  <= (OTHERS => '0');
          M_CBE_i(7 downto 0) <= "00000000";
          REQUEST64_i   <= '0';
          req64n      <= '1';

        WHEN irq_wait =>

          buf_full_irq <= '0'; -- OT

          REQUEST_i <= '0';
          req32n    <= '1';

          ADIO_i(63 downto 0)  <= (OTHERS => '0');
          REQUEST64_i   <= '0';
          req64n      <= '1';

        WHEN others =>

          buf_full_irq <= '0'; -- OT

      END CASE;
    END PROCESS;


    -------------------------------------------------------------------------------
    retry_counter:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    VARIABLE count : std_logic_vector(4 downto 0);
    BEGIN
      IF (rstn='0') THEN
        count := "00000";
      ELSIF ( clk'event AND clk='1') THEN
        IF (state = idle) THEN
          count := "00000";
        ELSIF ((state = read_req) OR (state=report_req) OR (state = req_bus)) THEN -- OT 18.02.05: Added states
          count := count + 1;
        ELSE
          count := count;
        END IF ;
      END IF;

      retry_count <= count;
    END PROCESS;

    -------------------------------------------------------------------------------
    timeout_counter:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    VARIABLE count : std_logic_vector(4 downto 0);
    BEGIN
      IF (rstn='0') THEN
        count := "00000";
      ELSIF ( clk'event AND clk='1') THEN
        IF (state = read_req) THEN
          count := "00000";
        ELSIF (state = rd_retry) THEN
          count := count + 1;
        ELSE
          count := count;
        END IF ;
      END IF;

      timeout_count <= count;
    END PROCESS;


    -------------------------------------------------------------------------------
    delay_counter:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    VARIABLE count : std_logic_vector(7 downto 0);
    BEGIN
      IF (rstn='0') THEN
        count := "00000000";
      ELSIF ( clk'event AND clk='1') THEN
        count := count + 1;
      ELSE
        count := count;
      END IF ;

    delay_count <= count;
    END PROCESS;


    -------------------------------------------------------------------------------
    pci_address_increment:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        pci_addr_cnt <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (req64n = '0') THEN
        	 IF (burst_continue='0') THEN
          	pci_addr_cnt <= dma_addr;
          ELSIF (retry_reg='1') THEN -- OT 17.02.05: Do not reset address counter if former transaction was terminated by the latency timer or ended with a retry
          	pci_addr_cnt <= pci_addr_cnt -16; -- Retry
          ELSIF ( (NOT (qword_cnt=0)) OR (NOT (qword_cnt=1)) ) THEN -- No disconnect handling if all the qwords were transferred
           pci_addr_cnt <= pci_addr_cnt -8; -- Latency timer
          ELSE
          	pci_addr_cnt <= pci_addr_cnt;
          END IF;
        ELSIF (M_SRC_EN = '1' AND state = mem_wr_data_phase) THEN
          pci_addr_cnt <= pci_addr_cnt + 8;
        ELSE
          pci_addr_cnt <= pci_addr_cnt;
        END IF;
      END IF;
    END PROCESS;


    -------------------------------------------------------------------------------
    store_next_pci_address:
    -------------------------------------------------------------------------------
--    PROCESS (clk, rstn)
--    BEGIN
--      IF (rstn = '0') THEN
--        next_pci_addr <= (OTHERS => '0');
--      ELSIF (clk'event and clk = '1') THEN
--        IF (state=pci_disc_wod) THEN
--          next_pci_addr <= pci_addr_cnt - 16;
--        ELSE
--          next_pci_addr <= next_pci_addr;
--        END IF;
--      END IF;
--    END PROCESS;
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        next_pci_addr <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (state=pci_disc) THEN
          IF (disc_wod_reg='1' OR (retry_reg='1')) THEN -- OT 17.02.05: Added retry condition
            next_pci_addr <= pci_addr_cnt - 16;
          ELSE
            next_pci_addr <= pci_addr_cnt - 8;
          END IF;
        ELSE
          next_pci_addr <= next_pci_addr;
        END IF;
      END IF;
    END PROCESS;
    -- Comment:
    -- data on PCI bus is one word behind l_adi;
    -- Due to the output register it is in our case 2 words,
    -- therefore we have to subtract 16 from the pci-address
    -- (on 64 bit systems one PCI cycle increments the address by 8)
    
    ---------------------------------------------------------------------------
    pci_disconnect_detect:
    ---------------------------------------------------------------------------

    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          disc_state <= disc_idle;
        ELSIF (clk'event AND clk='1') THEN
          CASE disc_state IS

            WHEN disc_idle =>
              IF (disc_wod_reg='1' and state=pci_disc) THEN
                disc_state <= disc_wod;
              ELSIF (retry_reg='1' and state=pci_disc) THEN -- OT 17.02.05: Added disc_retry
              		disc_state <= disc_retry_burst;
              ELSIF (disc_wd_reg='1') THEN
                disc_state <= disc_wd;
              ELSIF (TIME_OUT='1' AND (NOT (qword_cnt=0 OR qword_cnt=1 OR qword_cnt=x"10"))) THEN -- OT 17.02.05: Added disc_latency. OT 19.02.05: Don't care if the last word was already transferred
              		disc_state <= disc_latency;
              ELSE
                disc_state <= disc_idle;
              END IF;

            WHEN disc_wod =>
              IF (state=mem_wr_data_phase) THEN
                disc_state <= disc_wait1;
              ELSE
                disc_state <= disc_wod;
              END IF;

            WHEN disc_wd =>
              IF (state=mem_wr_data_phase) THEN
                disc_state <= disc_idle;
              ELSE
                disc_state <= disc_wd;
              END IF;

            WHEN disc_wait1 =>
              IF (M_SRC_EN = '1' AND state = mem_wr_data_phase) THEN
                disc_state <= disc_idle;
              ELSE
                disc_state <= disc_wait1;
              END IF;

            WHEN disc_latency => -- OT 17.02.05: Added disc_latency
              IF (state=mem_wr_data_phase) THEN
                disc_state <= disc_idle;
              ELSE
                disc_state <= disc_latency;
              END IF;

            WHEN disc_retry_burst => -- OT 17.02.05: Added disc_retry_burst
              IF (M_DATA='1' AND (state=mem_wr_data_phase OR state=report_data_out OR state=rd_data_phase)) THEN
                disc_state <= disc_idle;
              ELSE
                disc_state <= disc_retry_burst;
              END IF;

          END CASE;
        END IF;

      END PROCESS;

      burst_continue <= '0' WHEN ((disc_state = disc_idle)) ELSE '1';


    ------------------------------------------------------------------------------------
    -- local qword_cnt counter keeps tracks of number of qwords get transfered on the local side
    ------------------------------------------------------------------------------------
    PROCESS (clk, rstn, state, req32n, req64n, qword_cnt, M_SRC_EN)
    BEGIN
      IF (rstn = '0') THEN
        qword_cnt <= (OTHERS => '0');
      ELSIF (clk'event AND clk='1') THEN
        IF ((req32n = '0' OR req64n = '0') AND burst_continue='0') THEN -- load at request
          qword_cnt <= x"00000010";
        ELSIF ((req32n = '0' OR req64n = '0') AND burst_continue='1') THEN
        	 IF (disc_state=disc_latency) THEN
        	     qword_cnt <= qword_cnt + 1;
        	 ELSE
                     qword_cnt <= qword_cnt + 2;
                 END IF;
        ELSIF (M_SRC_EN = '1' AND state = mem_wr_data_phase) THEN
          qword_cnt <= qword_cnt - 1;
        ELSE
          qword_cnt <= qword_cnt;
        END IF;
      END IF;
    END PROCESS;



    ---------------------------------------------------------------------------
    -- Last data phase logic (generates 'COMPLETE')
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        last_state <= idle;
      ELSIF (clk'event AND clk = '1') THEN
        CASE last_state IS

          WHEN idle =>
            IF (req64n = '1') THEN                -- request sigaled, go to request state
              last_state <= mw_wait;
            ELSE
              last_state <= idle;
            END IF;

          WHEN mw_wait  =>
--            IF (qword_cnt = 3 AND lm_dxfrn = '0') THEN  -- for 64bit PCI-bus
--            IF (qword_cnt = 2 AND lm_dxfrn = '0') THEN  -- for 32bit PCI-bus
--            IF ((qword_cnt=3 AND M_SRC_EN='1' AND M_FAIL64='0') OR -- 64 bit transfer
--                (qword_cnt=2 AND M_SRC_EN='1' AND M_FAIL64='1')) THEN -- 32 bit transfer
            IF ((qword_cnt < 4) AND M_SRC_EN='1') THEN
              last_state <= last;
            ELSIF (data_phase_fall = '1') THEN               -- end of data phase, return to idle
              last_state <= idle;
            ELSE
              last_state <= mw_wait;
            END IF;

          WHEN last =>
            IF (M_DATA='0') THEN
               last_state <= idle;
            ELSE
               last_state <= last;
            END IF;

          WHEN others =>
          last_state <= idle;

        END CASE;
      END IF;
    END PROCESS;



    -----------------------------------------------------------
    -- Register M_DATA
    ------------------------------------------------------------
   PROCESS (clk, rstn, M_DATA)
   BEGIN
     IF (rstn = '0') THEN
       data_phase_reg <= '0';
     ELSIF (clk'event AND clk = '1') THEN
       data_phase_reg <= M_DATA;
     END IF;
   END PROCESS;

    -----------------------------------------------------------
    -- Register retries, disconnects and fatal errors
    ------------------------------------------------------------
   PROCESS (clk, rstn, CSR)
   BEGIN
   IF (rstn='0') THEN
       fatal_reg <= '0';
       retry_reg <= '0';
       disc_wd_reg <= '0';
       disc_wod_reg <= '0';
   ELSIF (clk'event AND clk = '1') THEN
         if (M_ADDR_N='0') then -- clear at beginning
                fatal_reg <= '0';
                retry_reg <= '0';
                disc_wd_reg <= '0';
                disc_wod_reg <= '0';
         elsif (M_DATA='1') then
         fatal_reg <= CSR(39) OR CSR(38);
         retry_reg <= CSR(36) AND (NOT M_DATA_VLD);
         disc_wod_reg <= CSR(36) AND M_DATA_VLD; -- Need to distinguish between retry and disc_wod
         disc_wd_reg <= CSR(37);
         end if;
   END IF;
   END PROCESS;

   data_phase      <= M_DATA; -- lm_tsr(3);
   data_phase_fall <= data_phase_reg AND NOT data_phase;

    M_WRDN          <=  '0' when ((state=read_req) OR (state=rd_addr_phase) OR (state=rd_data_phase)) else '1'; -- Controlling the direction of the transactions
    REQUESTHOLD     <= '0';

  END behavior;
