--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : ddl_if.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de>
-- Company    :
-- Platform   :
-------------------------------------------------------------------------------
-- Description:
--
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author    Description
--
-- 07.07.2004  1.0      tilsner   created
--=============================================================================
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;

ENTITY mem_if IS
  GENERIC (
    memif_addr    : integer RANGE 0 TO 255 := 0;
    memif_length  : integer RANGE 0 TO 255 := 4;
    memif_csr     : integer RANGE 0 TO 255 := 8;
    memif_trigger : integer RANGE 0 TO 255 := 12);
  PORT (
    clk       : IN std_logic;
    rstn      : IN std_logic;

    -- RAM interface
    addr      : OUT std_logic_vector(23 DOWNTO 0);  -- 2^24=16 MByte (max. 32bit RAM on RORC)
    d         : IN  std_logic_vector(31 DOWNTO 0);
    oe        : OUT std_logic;
    
    -- Data output
    data      : OUT std_logic_vector(31 DOWNTO 0);
    d_valid   : OUT std_logic;
    eoe_flag  : OUT std_logic;
    busy_n    : IN  std_logic;
    active    : OUT std_logic;
    
    -- config IF
    conf_addr : IN  std_logic_vector(7 DOWNTO 0);
    conf_in   : IN  std_logic_vector(31 DOWNTO 0);
    conf_out  : OUT std_logic_vector(31 DOWNTO 0);
    conf_we   : IN  std_logic    
  );

END mem_if;
  

ARCHITECTURE rtl OF mem_if IS

  TYPE fsm_state_type IS (idle, addr_phase, data_phase, busy, wait0, wait1, wait2, wait3, length_output);
  SIGNAL state : fsm_state_type;
  
  SIGNAL we           : std_logic_vector(3 DOWNTO 0);
  SIGNAL block_base   : std_logic_vector(31 DOWNTO 0);
  SIGNAL block_len    : std_logic_vector(31 DOWNTO 0);
  SIGNAL csr          : std_logic_vector(31 DOWNTO 0);
  SIGNAL addr_reg     : std_logic_vector(23 DOWNTO 0);
  SIGNAL word_counter : std_logic_vector(31 DOWNTO 0);
  SIGNAL start        : std_logic;
  
  BEGIN

    ---------------------------------------------------------------------------
    config_if:
    ---------------------------------------------------------------------------
    we(0)    <= '1' WHEN (conf_we='1' AND conf_addr = memif_addr)    ELSE '0';
    we(1)    <= '1' WHEN (conf_we='1' AND conf_addr = memif_length)  ELSE '0';
    we(2)    <= '1' WHEN (conf_we='1' AND conf_addr = memif_csr)     ELSE '0';
    we(3)    <= '1' WHEN (conf_we='1' AND conf_addr = memif_trigger) ELSE '0';
    conf_out <= block_base WHEN (conf_addr = memif_addr)   ELSE
                block_len  WHEN (conf_addr = memif_length) ELSE
                csr        WHEN (conf_addr = memif_csr)    ELSE
                (OTHERS => '0');
    
    -------------------------------------------------------------------------------
    ctrl_register:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        block_base <= (OTHERS => '0');
        block_len  <= (OTHERS => '0');          -- transfer length in 32bit words
        csr        <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (we(0) = '1') THEN
          block_base <= conf_in;
        END IF;
        IF (we(1) = '1') THEN
          block_len <= conf_in;
        END IF;
        IF (we(2) = '1') THEN
          csr <= conf_in;
        END IF;
      END IF;  
    END PROCESS;

    -- csr(0):  auto run
    -------------------------------------------------------------------------------
    address_increment:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN
        addr_reg     <= (OTHERS => '0');
        word_counter <= (OTHERS => '0');
      ELSIF (clk'event and clk = '1') THEN
        IF (state = idle) THEN
          addr_reg     <= block_base(23 DOWNTO 0);
          word_counter <= (OTHERS => '0');
        ELSIF (state = addr_phase OR state = data_phase OR state = wait0 OR state = wait1) THEN
          addr_reg     <= addr_reg + 4;
          word_counter <= word_counter + 1;
        END IF;
      END IF;
    END PROCESS;


    start <= csr(0) OR we(3);
    
    ---------------------------------------------------------------------------
    fsm:
    ---------------------------------------------------------------------------      
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          state <= idle;
        ELSIF rising_edge(clk) THEN
          CASE state IS

            WHEN idle =>
              IF (start='1') THEN
                state <= addr_phase;
              ELSE
                state <= idle;
              END IF;
              
            WHEN addr_phase =>
              state <= wait0;

            -- wait cycles are needed because of the delay between the address
            -- counter in this module and the appearence of the address at the
            -- output of the top_module
            WHEN wait0 =>
              state <=  wait1;
              
            WHEN wait1 =>
              state <=  data_phase;
                          
            WHEN data_phase =>
              IF (busy_n = '0') THEN
                state <= busy;
              ELSIF (word_counter = block_len) THEN
                state <= wait2;
              ELSE
                state <= data_phase;
              END IF;

            WHEN wait2 =>
              state <= wait3;

            WHEN wait3 =>
              state <= length_output;

            WHEN length_output =>
              state <= idle;
              
            WHEN busy =>
              IF (busy_n = '0') THEN
                state <= busy;
              ELSE
                state <= data_phase;
              END IF;
              
          END CASE;
        END IF;
      END PROCESS;


      -------------------------------------------------------------------------
      -- data path
      -------------------------------------------------------------------------
      active <= '1' WHEN (state /= idle) ELSE '0'; 
      oe     <= '0' WHEN (state /= idle) ELSE '1';
      addr   <= addr_reg;
      
      data     <= block_len WHEN (state=length_output) ELSE d;
      d_valid  <= '1'       WHEN (state=data_phase OR state=wait2 OR state=wait3 OR state=length_output) ELSE '0';
      eoe_flag <= '1'       WHEN (state=length_output) ELSE '0';
      
END rtl;
