--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : pattern_generator.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de>
-- Company    :
-- Platform   :
-------------------------------------------------------------------------------
-- Description:
--
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author      Description
--
-- 07.07.2004  1.0      tilsner     created
--
-- 09.03.2005  1.1      A. Szostak  conf_addr is no longer compared to
--                                  hardcoded values.
--=============================================================================


LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;

ENTITY pattern_generator IS

  GENERIC (
    patgen_evlen_reg  : integer RANGE 0 TO 255 := 0;
    patgen_csr_reg    : integer RANGE 0 TO 255 := 4;
    patgen_start_reg  : integer RANGE 0 TO 255 := 8
  );
  
  PORT (
    clk          : IN std_logic;
    rstn         : IN std_logic;

    -- Data interface
    data         : OUT std_logic_vector(31 DOWNTO 0);
    d_valid      : OUT std_logic;
    eoe_flag     : OUT std_logic;
    busy_n       : IN  std_logic;

    -- config IF
    conf_addr    : IN  std_logic_vector(7 DOWNTO 0);
    conf_in      : IN  std_logic_vector(31 DOWNTO 0);
    conf_out     : OUT std_logic_vector(31 DOWNTO 0);
    conf_we      : IN  std_logic
  );  

END pattern_generator;


ARCHITECTURE behaviour OF pattern_generator IS

  -- state machine encoding
  TYPE fsm_state_type IS (idle, header, data_wrd, status_wrd, busy1, busy2);
  SIGNAL state : fsm_state_type;

  SIGNAL we          : std_logic_vector(2 DOWNTO 0);
  SIGNAL ev_length   : std_logic_vector(18 DOWNTO 0);
  SIGNAL csr         : std_logic_vector(7 DOWNTO 0);
  SIGNAL pg_count    : std_logic_vector(31 DOWNTO 0);
  SIGNAL pg_rst      : std_logic;
  SIGNAL pg_run      : std_logic;
  SIGNAL ddl_headern : std_logic;
  
  SIGNAL header_wrd  : std_logic_vector(2 DOWNTO 0);
  SIGNAL bunch_cross : std_logic_vector(11 DOWNTO 0);
  SIGNAL orbit_number: std_logic_vector(23 DOWNTO 0);
  
  BEGIN
    
    ---------------------------------------------------------------------------
    config_if:
    ---------------------------------------------------------------------------
    we(0)    <= '1' WHEN (conf_we='1' AND conf_addr = patgen_evlen_reg) ELSE '0';
    we(1)    <= '1' WHEN (conf_we='1' AND conf_addr = patgen_csr_reg)   ELSE '0';
    we(2)    <= '1' WHEN (conf_we='1' AND conf_addr = patgen_start_reg) ELSE '0';
    conf_out <= x"000" & '0' & ev_length WHEN (conf_addr = patgen_evlen_reg) ELSE
                x"000000" & csr          WHEN (conf_addr = patgen_csr_reg)   ELSE
                (OTHERS => '0');
    

    ---------------------------------------------------------------------------
    ev_length_register:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          ev_length <= (OTHERS => '0');
        ELSIF (clk'event AND clk='1') THEN
          IF (we(0) = '1') THEN
            ev_length <= conf_in(18 DOWNTO 0);
          ELSE
            ev_length <= ev_length;
          END IF;
        END IF;
      END PROCESS;

      
    ---------------------------------------------------------------------------
    cs_register:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          csr <= (OTHERS => '0');
        ELSIF (clk'event AND clk='1') THEN
          IF (we(1) = '1') THEN
            csr <= conf_in(7 DOWNTO 0);
          ELSE
            csr <= csr;
          END IF;
        END IF;
      END PROCESS;

      pg_rst      <= csr(0);             -- reset internal pattern generator
      pg_run      <= csr(1);             -- free running
      ddl_headern <= csr(2);             -- w/o (0) (default) or w/ (1) DDL header
      
    ---------------------------------------------------------------------------
    data_gen:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          state <= idle;
        ELSIF (clk'event AND clk='1') THEN 
          CASE state IS
            WHEN idle =>
              IF (pg_run = '1' OR we(2)='1') THEN
                IF (ddl_headern = '0') THEN
                  state <= header;
                ELSE
                  state <= data_wrd;
                END IF;
              ELSE
                state <= idle;
              END IF;
              
            WHEN header =>
              IF (pg_rst = '1') THEN
                state <= idle;
              ELSIF (busy_n = '0') THEN
                state <= busy1;
              ELSIF (header_wrd = 7) THEN
                state <= data_wrd;
              ELSE
                state <= header;
              END IF;
              
            WHEN data_wrd =>
              IF (pg_rst = '1') THEN
                state <=  idle;
              ELSIF (busy_n = '0') THEN
                state <= busy2;
              ELSIF (pg_count = ev_length - 1) THEN
                state <= status_wrd;
              ELSE
                state <= data_wrd;
              END IF;

            WHEN status_wrd =>
              IF (pg_run = '1') THEN
                state <= header;
              ELSE
                state <= idle;
              END IF;

            WHEN busy1 =>
              IF (busy_n = '0') THEN
                state <= busy1;
              ELSE
                state <= header;
              END IF;
 
            WHEN busy2 =>
              IF (busy_n = '0') THEN
                state <= busy2;
              ELSE
                state <= data_wrd;
              END IF;
                
          END CASE;
        END IF;
      END PROCESS;

    -------------------------------------------------------------------------
    fsm_data:
    -------------------------------------------------------------------------
    PROCESS (state)
    BEGIN

      d_valid  <= '0';
      eoe_flag <= '0';
      
      CASE state IS
        WHEN idle =>
          d_valid  <= '0';
          eoe_flag <= '0';

        WHEN header =>
          d_valid  <= '1';
          eoe_flag <= '0';
          
        WHEN data_wrd =>
          d_valid  <= '1';
          eoe_flag <= '0';
            
        WHEN status_wrd =>
          d_valid  <= '1';
          eoe_flag <= '1';

        WHEN OTHERS =>
          d_valid  <= '0';
          eoe_flag <= '0';
         
      END CASE;
    END PROCESS;

    
    ---------------------------------------------------------------------------
    data_pattern_gen:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN                  -- asynchronous reset
        pg_count <= (OTHERS => '0');
      ELSIF (clk'event AND clk = '1') THEN
        IF (state=idle OR state=status_wrd) THEN
          pg_count <= (OTHERS => '0');
        ELSIF (state=data_wrd) THEN
          pg_count <= pg_count + 1;
        ELSE
          pg_count <= pg_count;      
        END IF;
      END IF;
    END PROCESS;


    ---------------------------------------------------------------------------
    bunch_cross_counter:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN                  -- asynchronous reset
        bunch_cross <= (OTHERS => '0');
      ELSIF (clk'event AND clk = '1') THEN
        IF (state=status_wrd) THEN
          bunch_cross <= bunch_cross + 1;
        ELSE
          bunch_cross <= bunch_cross;
        END IF;
      END IF;
    END PROCESS;

    
    ---------------------------------------------------------------------------
    orbit_counter:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN                  -- asynchronous reset
        orbit_number <= (OTHERS => '0');
      ELSIF (clk'event AND clk = '1') THEN
        IF (state=status_wrd) THEN
          orbit_number <= orbit_number + 1;
        ELSE
          orbit_number <= orbit_number;
        END IF;
      END IF;
    END PROCESS;
    
    ---------------------------------------------------------------------------
    header_wrd_counter:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn = '0') THEN                  -- asynchronous reset
        header_wrd <= (OTHERS => '0');
      ELSIF (clk'event AND clk = '1') THEN
        IF (state=status_wrd) THEN
          header_wrd <= (OTHERS => '0');
        ELSIF (state=header) THEN
          header_wrd <= header_wrd + 1;
        ELSE
          header_wrd <= header_wrd;
        END IF;
      END IF;
    END PROCESS;


    -- ouput signal with DDL header
    data <= x"ffffffff"                                  WHEN (state = header AND header_wrd = 0) ELSE
            x"01" & x"00" & x"0" & bunch_cross           WHEN (header_wrd = 1)     ELSE
            x"00" & orbit_number                         WHEN (header_wrd = 2)     ELSE
            -- add lenght of header (8 words)
            pg_count(31 DOWNTO 0) + 8                    WHEN (state=status_wrd AND ddl_headern='0') ELSE  
            pg_count(31 DOWNTO 0)                        WHEN (state=status_wrd AND ddl_headern='1') ELSE
            pg_count                                     WHEN (state=data_wrd)   ELSE (OTHERS => '0');


    
END behaviour;
