--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : pci2mem.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de>
-- Company    : 
-- Platform   : 
-------------------------------------------------------------------------------
-- Description:
--
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author    Description
--
-- 07.11.2004  1.0      tilsner   created
-- 13.01.2005  1.1      tilsner   SRAM64 interface added
--=============================================================================

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;

ENTITY pci2mem IS
  GENERIC (
    bar_mem32 : integer RANGE 0 TO 15 := 0;
    bar_mem64 : integer RANGE 0 TO 15 := 1);
  PORT (
    clk             : in std_logic;                                 
    rstn            : in std_logic; 

    -- PCI
    l_adro          : in std_logic_vector(63 downto 0);         
    l_cmdo          : in std_logic_vector(3 downto 0);          
    l_ldat_ackn     : in std_logic;
    l_hdat_ackn     : in std_logic;
    lt_abortn       : out std_logic;
    lt_discn        : out std_logic;                            
    lt_rdyn         : out std_logic;
    lt_framen       : in  std_logic;
    lt_ackn         : IN  std_logic;                            
    lt_dxfrn        : IN  std_logic;                           
    lt_tsr          : IN  std_logic_vector(11 downto 0);         

    --32bit SRAM interface
    sram32_a        : OUT std_logic_vector(23 DOWNTO 0);  -- 2^24 = 16 MByte (max. 32bit RAM on RORC) 
    sram32_we       : OUT std_logic;
    sram32_oe       : OUT std_logic;
        
    -- 64bit SRAM interface
    sram64_a        : OUT std_logic_vector(21 DOWNTO 0);
    sram64_oe       : OUT std_logic_vector(1 DOWNTO 0);
    sram64_we       : OUT std_logic_vector(1 DOWNTO 0);
    sram64_cs       : OUT std_logic;

    pci_out         : OUT std_logic
  );

END pci2mem;



ARCHITECTURE rtl OF pci2mem IS

  -- state machine encoding (one hot)
  TYPE state_type IS (idle, mem_wr32, mem_rd32addr, mem_rd32, mem_wr64, mem_rd64addr, mem_rd64, tranx_abort);
  SIGNAL state : state_type;

  SIGNAL successful_data_tranx : std_logic;
  SIGNAL bit64_tranx           : std_logic;
  SIGNAL burst_tranx           : std_logic;
  SIGNAL addr_inc              : std_logic_vector(23 DOWNTO 0);
  SIGNAL runWatchDog           : std_logic;
  SIGNAL abort                 : std_logic;
  
  BEGIN

    pci_out <= '1' WHEN (state=mem_rd32 OR state=mem_rd64) ELSE '0';

    -------------------------------------------------------------------------------
    WATCHDOG:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    VARIABLE count : std_logic_vector(4 downto 0);
    BEGIN 
      IF (rstn='0') THEN 
        count := "00000";
      ELSIF rising_edge(clk) THEN
        IF (state=idle) THEN 
          count := "00000";
        ELSIF (runWatchDog='1') THEN
          count := count + '1';
        END IF ;
      END IF;
    
      abort <= count(4);
    END PROCESS;
    runWatchDog <= '1' WHEN (state=mem_rd32 OR state=mem_wr32) ELSE '0';
    
    ---------------------------------------------------------------------------
    fsm_controller:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn, lt_framen, l_cmdo, state, lt_tsr, burst_tranx)
    BEGIN
      IF (rstn='0') THEN
        state <= idle;
      ELSIF rising_edge(clk) THEN
        CASE state IS
          WHEN idle =>
            -- memory write transaction to 32bit memory
            IF (lt_framen = '0' AND l_cmdo = "0111" AND lt_tsr(bar_mem32) = '1') THEN   
              state <= mem_wr32;

            -- memory write transaction to 64bit memory
            ELSIF (lt_framen = '0' AND l_cmdo = "0111" AND lt_tsr(bar_mem64) = '1') THEN
              state <= mem_wr64;
   
            -- memory read transaction to 32bit memory
            ELSIF (lt_framen = '0' and l_cmdo = "0110" and lt_tsr(bar_mem32) = '1' and burst_tranx = '1') THEN 
              state <= mem_rd32addr;
 
            ELSIF (lt_framen = '0' and l_cmdo = "0110" and lt_tsr(bar_mem32) = '1' and burst_tranx = '0') THEN
              state <= mem_rd32addr;

            -- memory read transaction to 64bit memory (bar 1)
            ELSIF (lt_framen = '0' and l_cmdo = "0110" and lt_tsr(bar_mem64) = '1' and burst_tranx = '1') THEN
               state <= mem_rd64addr;
               
            ELSIF (lt_framen = '0' and l_cmdo = "0110" and lt_tsr(bar_mem64) = '1' and burst_tranx = '0') THEN
               state <= mem_rd64addr;
               
            ELSE
              state <= idle;
            END IF;

                      
          WHEN mem_wr32 =>
            IF (abort = '1') THEN
              state <= tranx_abort;
            ELSIF (lt_framen = '1') THEN
              state <= idle;
            ELSE
              state <= mem_wr32;
            END IF;

            
          WHEN mem_rd32addr =>
            state <= mem_rd32;

            
          WHEN mem_rd32 =>
            IF (abort = '1') THEN
              state <= tranx_abort;             
            ELSIF (lt_framen = '1') THEN
              state <= idle;
            ELSE
              state <= mem_rd32;
            END IF;            
            
            
          WHEN mem_wr64 =>
            IF (lt_framen = '1') THEN
              state <= idle;
            ELSE
              state <= mem_wr64;
            END IF;

            
          WHEN mem_rd64addr =>
            state <= mem_rd64;

            
          WHEN mem_rd64 =>
            IF (lt_framen = '1') THEN
              state <= idle;
            ELSE
              state <= mem_rd64;
            END IF;


          WHEN tranx_abort =>
            IF (lt_framen = '1') THEN 
              state <= idle;
            ELSE
              state <= tranx_abort;
            END IF;
            
            
          WHEN OTHERS =>
            state <= idle;
            
        END CASE;
      END IF;        
    END PROCESS;
      

    -- There is a successful data transfer when lt_dxfrn(local target data transfer)
    -- and lt_ackn(local target acknowldge) are asserted at the same time.
    -- A 64 bit burst transfer is indicated by lt_tsr(7)(transfer 64) and 
    -- lt_tsr(9)(burst transfer).
    successful_data_tranx <= NOT(lt_dxfrn) AND NOT(lt_ackn); 
    burst_tranx <= lt_tsr(9); 
    bit64_tranx <= lt_tsr(7);

    ---------------------------------------------------------------------------
    data_controller:
    ---------------------------------------------------------------------------
    PROCESS (clk, state, l_ldat_ackn, l_hdat_ackn,
             successful_data_tranx, lt_framen, bit64_tranx, burst_tranx)
    BEGIN
      -- default values
      lt_rdyn   <= '1';
      lt_abortn <= '1';
      lt_discn  <= '1';
      sram32_we <= '1';
      sram32_oe <= '1';
      sram64_we <= "11";
      sram64_oe <= "11";
      sram64_cs <= '1';
      
      CASE state IS
        WHEN idle =>
          -- initialization
          -- local target control signals
          
          sram32_we <= '1';
          sram32_oe <= '1';

          sram64_we <= "11";
          sram64_oe <= "11";
          sram64_cs <= '1';
          
        WHEN mem_wr32 =>
          -- ready to accept data
          lt_rdyn   <= '0';

--          IF (burst_tranx='1') THEN
--            lt_discn <= '0';
--          ELSE
--            lt_discn <= '1';
--          END IF;

          sram32_we <= NOT ((NOT(l_ldat_ackn) OR NOT(l_hdat_ackn)) AND successful_data_tranx);
          sram32_oe <= '1';

        WHEN mem_rd32addr =>
          lt_rdyn   <= '1';
          sram32_oe <= '0';
          
        WHEN mem_rd32 =>
          -- ready to provide data        
          lt_rdyn   <= '0';
          sram32_oe <= '0';

--          IF (burst_tranx='1' AND successful_data_tranx='1') THEN
--            lt_discn <= '0';
--          ELSE
--            lt_discn <= '1';
--          END IF;

          
        WHEN mem_wr64 =>
          -- ready to accept data
          lt_rdyn      <= '0';
          
          IF burst_tranx='1' THEN
            lt_discn <= '0';
          ELSE
            lt_discn <= '1';
          END IF;

          sram64_we(0) <= NOT(NOT(l_ldat_ackn) AND successful_data_tranx);
          sram64_we(1) <= NOT(NOT(l_hdat_ackn) AND successful_data_tranx);
          sram64_oe    <= "11";
          sram64_cs    <= '0';
          
        WHEN mem_rd64addr =>
          lt_rdyn   <= '1';
          sram64_oe <= "00";
          sram64_cs <= '0';
          
        WHEN mem_rd64 =>
          -- read< to provide data
          lt_rdyn      <= '0';
          sram64_oe    <= "00";
          sram64_cs    <= '0';
          
          IF (burst_tranx='1') THEN
            lt_discn <= '0';
          ELSE
            lt_discn <= '1';
          END IF;
          
        WHEN tranx_abort =>
          lt_abortn <= '0';
          
        WHEN others => 
          lt_rdyn <= '1';

      END CASE;
    END PROCESS;

    -------------------------------------------------------------------------------
    address_counter:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF rstn = '0' THEN                     -- asynchronous reset (active low)
        addr_inc <= (OTHERS => '0');
      ELSIF rising_edge(clk) THEN
        IF (state = idle) THEN
          addr_inc <= l_adro(23 DOWNTO 0);
        ELSIF (successful_data_tranx = '1') THEN
          addr_inc <= addr_inc + 4;
        END IF;
      END IF;
    END PROCESS;
    


    sram32_a     <= addr_inc;
    sram64_a     <= l_adro(21 DOWNTO 0);
    
  END rtl;
