--=============================================================================
-- Project    : HLT-RORC
-------------------------------------------------------------------------------
-- File       : target_ctrl.vhd
-- Author     : Heinz Tilsner  <tilsner@kip.uni-heidelberg.de> / Olav Torheim <torheim@ift.uib.no>
-- Company    :
-- Platform   :
-------------------------------------------------------------------------------
-- Description:
--
--
-------------------------------------------------------------------------------
-- Revision list:
-- Date        Version  Author    Description
--
-- 07.07.2004  1.0      tilsner   created
--
-- 24.06.2005  1.1      O. Torheim  Replacing the Altera MegaCore interface with the Xilinx LogiCore interface
--=============================================================================

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;

ENTITY target_ctrl IS

  PORT (
    -- PCI interface
    clk           : IN std_logic;
    rstn          : IN std_logic;
    ADDR            : in    std_logic_vector( 31 downto 0);
    ADIO            : inout std_logic_vector( 63 downto 0);
    BASE_HIT        : in    std_logic_vector(  7 downto 0);
    S_ABORT         : out   std_logic;
    S_WRDN          : in    std_logic;
    IDLE            : in    std_logic;
    S_DATA_VLD      : in    std_logic;

    -- interface to logic
    pci_out       : OUT std_logic;

    -- configuration interface
    conf_rd       : IN  std_logic_vector(31 DOWNTO 0);
    conf_wr       : OUT std_logic_vector(31 DOWNTO 0);
    conf_addr     : OUT std_logic_vector(7 DOWNTO 0);
    conf_we       : OUT std_logic
  );

END target_ctrl;



ARCHITECTURE behaviour OF target_ctrl IS

   -- state machine encoding
  TYPE fsm_state_type IS (pci_idle, pci_write, pci_read, pci_abort);
  SIGNAL fsm_state : fsm_state_type;

  SIGNAL S_ABORT_reg  : std_logic;
  SIGNAL S_ABORT_i    : std_logic;
--  SIGNAL lt_discn_reg   : std_logic;
--  SIGNAL lt_discn_i     : std_logic;
--  SIGNAL burst_tranx    : std_logic;
  SIGNAL runWatchDog  : std_logic; -- Changed name. OT 27.05.05
  SIGNAL abort          : std_logic;
  SIGNAL conf_addr_reg  : std_logic_vector(7 DOWNTO 0);
  SIGNAL conf_wr_reg    : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_rd_reg    : std_logic_vector(31 DOWNTO 0);
  SIGNAL conf_we_reg    : std_logic;

   -- New inserted signals:
  signal oe_io_reg      : std_logic;

  BEGIN  -- behaviour

    pci_out <= '1' WHEN (fsm_state=pci_read) ELSE '0';

    ---------------------------------------------------------------------------
    pci_output_regs:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
      BEGIN
        IF (rstn='0') THEN
          S_ABORT_reg <= '0';
--          lt_discn_reg  <= '1';
        ELSIF (clk'event AND clk='1') THEN
          S_ABORT_reg <= S_ABORT_i;
--          lt_discn_reg  <= lt_discn_i;
        END IF;
    END PROCESS;
    S_ABORT <= S_ABORT_reg;
--    lt_discn  <= lt_discn_reg;

    -------------------------------------------------------------------------------
    WATCHDOG:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    VARIABLE count : std_logic_vector(4 downto 0);
    BEGIN
      IF (rstn='0') THEN
        count := "00000";
      ELSIF ( clk'event AND clk='1') THEN
        IF (runWatchDog='0') THEN
          count := "00000";
        ELSE
          count := count + '1';
        END IF ;
      END IF;
          abort <= count(4);
    END PROCESS;

    runWatchDog <= '1' WHEN (fsm_state=pci_read OR fsm_state=pci_write) ELSE '0';

    -------------------------------------------------------------------------------
    fsm_controller:
    -------------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn='0') THEN
        fsm_state <= pci_idle;
      ELSIF (clk'event AND clk='1') THEN
        CASE fsm_state IS

          WHEN pci_idle =>
            --memory write transaction
--            IF (lt_framen = '0' AND l_cmdo = "0111" AND lt_tsr(0) = '1') THEN
	      IF (S_WRDN = '1' AND BASE_HIT(0) = '1') THEN
              fsm_state <= pci_write;
            ELSIF (S_WRDN = '0' AND BASE_HIT(0) = '1') THEN
              fsm_state <= pci_read;
            ELSE
              fsm_state <= pci_idle;
            END IF;

          WHEN pci_write =>
            IF (abort='1') THEN
              fsm_state <= pci_abort;
            ELSIF (IDLE = '1') THEN
              fsm_state <= pci_idle;
            ELSE
              fsm_state <= pci_write;
            END IF;

          WHEN pci_read =>
            IF (IDLE = '1') THEN
              fsm_state <= pci_idle;
            ELSE
              fsm_state <= pci_read;
            END IF;

          WHEN pci_abort =>
            fsm_state <= pci_idle;

        END CASE;
      END IF;
    END PROCESS;


    -------------------------------------------------------------------------------
    fsm_data_controller:
    -------------------------------------------------------------------------------
    PROCESS (fsm_state, conf_rd_reg)
    BEGIN

--      lt_discn_i  <= '1';

      CASE fsm_state IS

        WHEN pci_idle =>
	  oe_io_reg <= '0';
          S_ABORT_i <= '0';

        WHEN pci_write =>
	  oe_io_reg <= '0';
          S_ABORT_i <= '0';
--          IF (burst_tranx='1') THEN
--            lt_discn_i <= '0';
--          ELSE
--            lt_discn_i <= '1';
--          END IF;

        WHEN pci_read =>
	  oe_io_reg <= '1';
          S_ABORT_i <= '0';

--          IF (burst_tranx='1') THEN
--            lt_discn_i <= '0';
--          ELSE
--            lt_discn_i <= '1';
--          END IF;

        WHEN pci_abort =>
          S_ABORT_i <= '1';
	  oe_io_reg <= '0';

        WHEN others =>
          S_ABORT_i <= '0';
	  oe_io_reg <= '0';
      END CASE;
    END PROCESS;



    ---------------------------------------------------------------------------
    conf_address_register:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn='0') THEN
        conf_addr_reg <= (OTHERS => '0');
      ELSIF (clk'event AND clk='1') THEN
        IF (fsm_state = pci_idle) THEN
          conf_addr_reg <= ADDR(7 DOWNTO 0);
        ELSE
          conf_addr_reg <= conf_addr_reg;
        END IF;
      END IF;
    END PROCESS;

    ---------------------------------------------------------------------------
    conf_data_out_register:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn='0') THEN
        conf_wr_reg <= (OTHERS => '0');
      ELSIF (clk'event AND clk='1') THEN
        IF (fsm_state=pci_write AND S_DATA_VLD='1') THEN
          conf_wr_reg <= ADIO(31 DOWNTO 0);
        ELSE
          conf_wr_reg <= conf_wr_reg;
        END IF;
      END IF;
    END PROCESS;

    ---------------------------------------------------------------------------
    conf_data_in_register:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn='0') THEN
        conf_rd_reg <= (OTHERS => '0');
      ELSIF (clk'event AND clk='1') THEN
        conf_rd_reg <= conf_rd;
      ELSE
        conf_rd_reg <= conf_rd_reg;
      END IF;
    END PROCESS;

    ---------------------------------------------------------------------------
    conf_we_register:
    ---------------------------------------------------------------------------
    PROCESS (clk, rstn)
    BEGIN
      IF (rstn='0') THEN
        conf_we_reg <= '0';
      ELSIF (clk'event AND clk='1') THEN
        IF (fsm_state = pci_idle) THEN
          conf_we_reg <= '0';
        ELSIF (fsm_state=pci_write and S_DATA_VLD='1') THEN
          conf_we_reg <= '1';
        ELSE
          conf_we_reg <= '0';
        END IF;
      END IF;
    END PROCESS;

--  ADIO(31 downto 0) <= conf_rd_reg when (oe_io_reg = '1') else (others=>'Z');
  ADIO(31 downto 0) <= conf_rd when (oe_io_reg = '1') else (others=>'Z');

    -- output signals
    conf_addr <= conf_addr_reg;
    conf_wr   <= conf_wr_reg;
    conf_we   <= conf_we_reg;

END behaviour;
