/*********************************************************************************

   Author: Olav Torheim
   Email:  torheim@ift.uib.no

   A kernel module implementing a simple driver for HLT-RORC.
   Parts of this code come from the irqsignal module created by Artur Szostak (artur@alice.phy.uct.ac.za).

**********************************************************************************/



#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/config.h>

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/delay.h>  /* udelay */
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/tqueue.h>
#include <linux/spinlock.h>
#include <linux/pci.h>

#include <linux/wrapper.h>

#include <asm/semaphore.h> /* Needed for semaphores: */

#include <asm/io.h>
#include <asm/uaccess.h>

#include "rorcdriver_mod.h"

#ifdef WITHBIGPHYS
#include <linux/bigphysarea.h>
#endif

#ifndef KERNEL_VERSION /* pre-2.1.90 didn't have it */
#  define KERNEL_VERSION(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif

/* Only support kernels newer than 2.4.x */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#  error "This kernel is too old: not supported by this file"
#endif


// Some module descriptions...
MODULE_AUTHOR("Olav Torheim");
MODULE_DESCRIPTION("Simple driver for HLT-RORC.");
MODULE_LICENSE("GPL");


/* The following parameters can be passed in during module load time.
     eg. insmod rorcdriver major=36
 */
static int major = 0; // Major version of character device.
MODULE_PARM(major, "i");

spinlock_t irqcounter_lock = SPIN_LOCK_UNLOCKED;

int irqcounter;

struct pci_dev *pdev; // Pointer to PCI device structure
char * baseptr; // Virtual base address pointer for BAR0 of HLT-RORC
int pg_en;

unsigned long event_buffer_length; // Event buffer length in words
unsigned long report_buffer_length; // Report buffer length in words

unsigned long report_buffer_counter;
unsigned long report_buffer_processing_counter;

unsigned long dma_base_address; // PCI bus base address of event buffer
unsigned long read_ptr;

u32 * event_buffer; // Virtual base address of event buffer

unsigned long dma_report_address;

u32 * report_buffer; // Virtual base address of report buffer

u32 event_pages;
u32 report_pages;

struct page *page;

struct semaphore mr_sem; // Counting semaphore for synchronizing reading and writing in the event buffer
struct semaphore rd_ptr_mutex; // Mutex semaphore for synchronizing read pointer updating

struct timer_list test_timer;

int timeout_in_cycles;
int timeout_in_system_time;

const long errorcode = -1;


/* This routine actually performs the interrupt handling after being
   scheduled by the interrupt service routine.
 */
void rorcdriver_dispatcher(void* _ireg /* read as: u32 interrupt_status_reg */)
{

	u32 interrupt_status_reg = (u32)_ireg;

	int irqcounter_local, counter_local;

	unsigned long flags;
	spin_lock_irqsave(&irqcounter_lock, flags); // Using spin_lock_irqsave to prevent from races between ISR and dispatcher routine
		irqcounter_local = irqcounter;
		irqcounter = 0;
	spin_unlock_irqrestore(&irqcounter_lock, flags);

	for(counter_local=0;counter_local<irqcounter_local;counter_local++)
	{
		up(&mr_sem); // Release sempahore if a new event has arrived
	}

//       if ((interrupt_status_reg==3) || (interrupt_status_reg==2)) // Release sempahore if a new event has arrived
//              {
//         up(&mr_sem);
//       }

       if ((interrupt_status_reg==3) || (interrupt_status_reg==1)) // Write back new read pointer value
	{
                down_interruptible(&rd_ptr_mutex); // Wait for read pointer to be updated
                writel((dma_base_address + read_ptr),baseptr+0x30); // Writing back read pointer
	}

        if (pg_en==1) // If pattern generation is enabled
        {
           writel(0xAFFED00F, baseptr+0x28); // Tell HLT-RORC to generate another event.
        }


};


/* Setup the task structure for the rorcdriver_dispatcher routine.
   This task will be scheduled by the interrupt handler: rorcdriver_interrupt
 */
struct tq_struct rorcdriver_dispatcher_task =
{
	LIST_HEAD_INIT(rorcdriver_dispatcher_task.list),
	0,
	&rorcdriver_dispatcher,
	NULL
};


/* The interrupt service routine schedules the dispatcher to service the request from the device.
 */

u32 ireg;
void rorcdriver_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
#	ifdef DEBUG
	printk(KERN_INFO "%s: INTERRUPT START, irq %d\n", THIS_MODULE->name, irq);
#	endif

  if ((ireg = readl(baseptr+0x54)) ==0) // Not our interrupt
  	return;

	if (ireg >> 1) irqcounter++;
	rorcdriver_dispatcher_task.data = (void*)ireg; // Send the interrupt status register as input data to the dispatcher.
	schedule_task(&rorcdriver_dispatcher_task);

#	ifdef DEBUG
	printk(KERN_INFO "%s: INTERRUPT END, irq %d\n", THIS_MODULE->name, irq);
#	endif
};


static int rorcdriver_mmap(struct file *file, struct vm_area_struct *vma)
{
        vma->vm_flags |= VM_LOCKED|VM_SHARED;

        remap_page_range(vma, vma->vm_start, virt_to_phys(event_buffer), (event_buffer_length << 1), PAGE_SHARED); // Shifting left for doubling the buffer
        return 0;
}

/* Registers the current process with the driver.

   Possible return values:
     RORCDRIVER_OK                 - Registered successfully.
     RORCDRIVER_OUTOFMEM           - Could not allocate memory for new task entry.
                                 i.e. irq number >= MAX_NUMBER_OF_IRQS
     RORCDRIVER_UNKNOWN_ERROR      - "request_irq" returned a unexpected result.
     -EFAULT                   - Pointer (_info) to data in user space memory is invalid.
 */
rorcdriver_status_t rorcdriver_register(rorcdriver_register_info_t* _info)
{
	rorcdriver_register_info_t info;
	int result;

	unsigned long flags;
	spin_lock_irqsave(&irqcounter_lock, flags);
		irqcounter = 0;
	spin_unlock_irqrestore(&irqcounter_lock, flags);

	// We need to copy data from user space into kernel space.
	if ( copy_from_user(&info, _info, sizeof(rorcdriver_register_info_t)) )
		return -EFAULT;

		result = request_irq(pdev->irq, rorcdriver_interrupt, SA_INTERRUPT | SA_SHIRQ, THIS_MODULE->name, pdev);
		if (result != 0) goto handle_error;

        timeout_in_cycles = ((info.timeout_in_millisecs * HZ) / 1000); // Convert from millisecs to clock cycles

	sema_init(&mr_sem, 0); // Initializing the end of event counting semaphore
        init_MUTEX_LOCKED(&rd_ptr_mutex);

        pg_en = info.enable_pg_transfers;
        event_buffer_length = info.event_buffer_length;
        report_buffer_length = info.report_buffer_length;
        report_buffer_counter = 0;
        report_buffer_processing_counter = 0;

        #ifdef WITHBIGPHYS
        int byte_count;
        byte_count = event_buffer_length << 2; // Number of bytes in double segmented buffer
        event_pages = byte_count / PAGE_SIZE;
        if ((byte_count % PAGE_SIZE) != 0) event_pages++;

        byte_count = report_buffer_length << 1; // Twice as many bytes as there are words
        report_pages = byte_count / PAGE_SIZE;
        if ((byte_count % PAGE_SIZE) != 0) report_pages++;

        event_buffer = (u32 *)bigphysarea_alloc_pages(event_pages,1,GFP_KERNEL); // Eventbuffer allocated for DMA transfers. Doubling the event buffer for getting hold of a region to place processed data

        page = virt_to_page(event_buffer);
        mem_map_reserve(page); // Page has to be reserved before becoming accessible from remap_page_range()

        report_buffer = (u32 *)bigphysarea_alloc_pages(report_pages,1,GFP_KERNEL); // Report buffer

        #else
        event_buffer = kmalloc((event_buffer_length << 1),  GFP_DMA); // Eventbuffer allocated for DMA transfers. Doubling the event buffer for getting hold of a region to place processed data
        page = virt_to_page(event_buffer);
        mem_map_reserve(page); // Page has to be reserved before becoming accessible from remap_page_range()

        report_buffer = kmalloc((sizeof(report_buffer)*report_buffer_length),  GFP_DMA); // Report buffer
        #endif

        dma_base_address = virt_to_bus(event_buffer); // Convert virtual memory address to PCI bus address
        dma_report_address = virt_to_bus(report_buffer); // Convert virtual memory address to PCI bus address

	// Determining values for the local control registers of HLT-RORC
        writel(dma_base_address, baseptr); // Base address for the DMA event buffer
        writel(info.event_buffer_length, baseptr+0x04);
        writel(dma_report_address, baseptr+0x08); // Base address for the DMA report buffer
        writel((dma_base_address + read_ptr),baseptr+0x30); // Value for read pointer
        writel(info.irq_pulse_length, baseptr+0x58);

        writel(info.report_buffer_length, baseptr+0x34); // Length of report buffer

        unsigned long global_csr = 0;
        if (info.enable_dma_transfers == 1) global_csr = global_csr | 1; // DMA enable bit
        if (info.enable_pg_transfers == 1) global_csr = global_csr | 16; // Pattern Generator enable bit
        if (info.enable_irq_transfers == 1) global_csr = global_csr | 128; // IRQ enable bit

        writel(global_csr, baseptr+0x18);

        if (pg_en==1)
        {
                writel(info.event_length, baseptr+0x20);
                writel(0x1, baseptr+0x24); // Reset Pattern Generator
                writel(0x0, baseptr+0x24); // Free Pattern Generator

                writel(0xAFFED00F, baseptr+0x28); // Free Pattern Generator
        }

	return 0;

handle_error:

	switch (result)
	{
	case -ENOMEM: return RORCDRIVER_OUTOFMEM;
	default:      return RORCDRIVER_UNKNOWN_ERROR;
	};
};


/* Unregisters the current process from the rorcdriver. */
rorcdriver_status_t rorcdriver_unregister(rorcdriver_register_info_t* _info)
{

        free_irq(pdev->irq, pdev); // Free interrupt routine
        up(&rd_ptr_mutex); // Wake up sleeping dispatcher routine
        flush_scheduled_tasks(); // Wait for dispatcher routine to complete
        pg_en = 0; // Turn off pattern generation

        #ifdef WITHBIGPHYS
        mem_map_unreserve(page);
        bigphysarea_free_pages((caddr_t)event_buffer);
        bigphysarea_free_pages((caddr_t)report_buffer);
        #else
        mem_map_unreserve(page);
        kfree(event_buffer);
        kfree(report_buffer);
        #endif

	return RORCDRIVER_OK;
};


/*
   This function is left blank for eventual future implementations.
 */
rorcdriver_status_t rorcdriver_reset(unsigned int irq)
{

	return RORCDRIVER_OK;
};


/*
   This function is left blank for eventual future implementations.
 */
rorcdriver_status_t rorcdriver_resetall(void)
{
	return RORCDRIVER_OK;
};

int tid;

void rorcdriver_wakeupcall(){
	/* send the signal to bring the task out of down call*/
	kill_proc(tid, SIGALRM, 1);
}

rorcdriver_status_t rorcdriver_getreportword(u32 * report_word)
{
        tid = current->pid;

        timeout_in_system_time = timeout_in_cycles + jiffies;

	init_timer(&test_timer);
	test_timer.data=0;
	test_timer.expires = timeout_in_system_time;
	test_timer.function=rorcdriver_wakeupcall;

	add_timer(&test_timer);

        if (down_interruptible(&mr_sem)) // Waiting until there are report words ready
        {
            copy_to_user(report_word, &errorcode, sizeof(errorcode));
        }
        else
        {
            copy_to_user(report_word, &report_buffer[report_buffer_counter], sizeof(&report_buffer[report_buffer_counter]));
            report_buffer_counter++;
        }
	del_timer(&test_timer);

            if (report_buffer_counter == report_buffer_length) // Start over again if reaching end of report buffer
            {
             report_buffer_counter = 0;
            }


        return RORCDRIVER_OK;
};

int multiplication_factor;
int ddl_datalength;

rorcdriver_status_t rorcdriver_processingcompleted(void)
{
        if (read_ptr==(event_buffer_length - 256))
        {
          read_ptr = 0;
        }
        else
        {
         ddl_datalength = (report_buffer[report_buffer_processing_counter] >> 13) +1; // right shift to obtain length of DDL datastream
         multiplication_factor = ddl_datalength / 32;
         if (ddl_datalength % 32) multiplication_factor++;
         multiplication_factor = multiplication_factor << 2; // left shift to multiply by 4

         read_ptr = read_ptr + (32*multiplication_factor); // Increment read pointer
        }

            report_buffer_processing_counter++;
            if (report_buffer_processing_counter == report_buffer_length) // Start over again if reaching end of report buffer
            {
             report_buffer_processing_counter = 0;
            }

         up(&rd_ptr_mutex);

        return RORCDRIVER_OK;
};


/* Used to recover from serious problems during debugging.
 */
void rorcdriver_hardreset(void)
{
	rorcdriver_resetall();
	while (MOD_IN_USE) MOD_DEC_USE_COUNT;
	MOD_INC_USE_COUNT;  // leave the use count at 1 since at least the calling process
	                    // still has a file handle.
};


/* This routine handles opening of the device file.
 */
int rorcdriver_open(struct inode *inode, struct file *filp)
{
	unsigned int minor = MINOR(inode->i_rdev);
	printk(KERN_INFO "%s: Open for minor device %d.\n", THIS_MODULE->name, minor);

	return 0;
};


/* This routine handles closing of the device file.
   We cleanup after the current process if it didn't already do so with ioctl calls
   to unregister itself.
 */
int rorcdriver_release(struct inode *inode, struct file *filp)
{

	unsigned int minor = MINOR(inode->i_rdev);
	printk(KERN_INFO "%s: Release for minor device %d.\n", THIS_MODULE->name, minor);

	free_irq(pdev->irq, pdev);

	return 0;
};


/* ioctl routine maps the ioctl numbers for this kernel module to the appropriate
   handling routines.
 */
int rorcdriver_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	rorcdriver_status_t status;
	unsigned int minor = MINOR(inode->i_rdev);

	switch (cmd)
	{
	case rorcdriver_REGISTER:
		printk (KERN_INFO "%s: ioctl rorcdriver_REGISTER for minor device %d.\n", THIS_MODULE->name, minor);
		status = rorcdriver_register( (rorcdriver_register_info_t*)arg );
    	break;

	case rorcdriver_UNREGISTER:
		printk (KERN_INFO "%s: ioctl rorcdriver_UNREGISTER for minor device %d.\n", THIS_MODULE->name, minor);
		status = rorcdriver_unregister( (rorcdriver_register_info_t*)arg );
    	break;

	case rorcdriver_RESET:
		printk (KERN_INFO "%s: ioctl rorcdriver_RESETIRQ for minor device %d.\n", THIS_MODULE->name, minor);
		status = rorcdriver_reset( (unsigned int)arg );
    	break;

	case rorcdriver_RESETALL:
		printk (KERN_INFO "%s: ioctl rorcdriver_RESETIRQALL for minor device %d.\n", THIS_MODULE->name, minor);
		status = rorcdriver_resetall();
    	break;

	case rorcdriver_HARDRESET:
		printk (KERN_INFO "%s: ioctl rorcdriver_HARDRESET for minor device %d.\n", THIS_MODULE->name, minor);
		rorcdriver_hardreset();
		status = RORCDRIVER_OK;
    	break;

	case rorcdriver_GETREPORTWORD:
		printk (KERN_INFO "%s: ioctl rorcdriver_GETREPORTWORD for minor device %d.\n", THIS_MODULE->name, minor);
		status = rorcdriver_getreportword( (u32*)arg );
    	break;

	case rorcdriver_PROCESSINGCOMPLETED:
		printk (KERN_INFO "%s: ioctl rorcdriver_TRANSFERCOMPLETED for minor device %d.\n", THIS_MODULE->name, minor);
		status = rorcdriver_processingcompleted();
    	break;

	default:
		printk (KERN_ERR "%s: Bad ioctl for device major number %d, minor number %d.\n", THIS_MODULE->name, major, minor);
		status = RORCDRIVER_NO_SUCH_IOCTRL;
	};
	return status;
};


/* Map the file operations to our routines.
 */
struct file_operations rorcdriver_fileops =
{
	ioctl:   rorcdriver_ioctl,
	mmap:    rorcdriver_mmap,
	open:    rorcdriver_open,
	release: rorcdriver_release,
	owner: THIS_MODULE,
};


/* Module initialisation:
   Allocate resources and register as a character device driver.
 */
int rorcdriver_init(void)
{
	int result;

	pdev = pci_find_device(0x10DC,0x033, NULL); // Assigning the HLT-RORC to the pci_dev data structure
	if (pdev == NULL)
	{
		printk(KERN_ERR "%s: Can't find hardware.\n", THIS_MODULE->name);
		return (-1);
	}

	baseptr = ioremap(pdev->resource[0].start, 1024); // Maps BAR0 high memory into kernel space

	printk (KERN_INFO "%s: Initialize.\n", THIS_MODULE->name);

	// Registering as acharacter device.
	// If the major number is zero then we will auto assign a major number.
	result = register_chrdev(major, "rorcdriver", &rorcdriver_fileops);
	if (result < 0)
	{
		printk(KERN_ERR "%s: Can't get major number.\n", THIS_MODULE->name);
		return result;
	}
	if (major == 0) major = result; /* dynamic */

	return 0;
}


/* Module cleanup:
   Unregister and release all resources.
 */
void rorcdriver_cleanup(void)
{
	iounmap(baseptr); // Unmaps BAR0 Memory Space

	printk (KERN_INFO "%s: Cleanup.\n", THIS_MODULE->name);
//	rorcdriver_resetall();  // Hope everyone released their resources.
	unregister_chrdev(major, "rorcdriver");
}


module_init(rorcdriver_init);
module_exit(rorcdriver_cleanup);

