/*********************************************************************************

   Author: Olav Torheim
   Email:  torheim@ift.uib.no

   Implementation of the rorcdriver API library.
   Parts of this code come from the irqsignal API created by Artur Szostak (artur@alice.phy.uct.ac.za).

**********************************************************************************/

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/mman.h> // Needed for memory mapping

#include "../../include/rorcdriver.h"

#define DRIVER_NODE "/dev/rorcdriver"


int drivernode = -1;  // file handle to the driver.
int registered_exit_routine = 0;  // Boolean value specifying if we registered the exit routine or not.

int * mmap_address;
int mmap_length;

void request_timeout_handler(int signal) // request_timeout_handler catches the SIGALRM signal generated by the rorcdriver kernel module
{
// Do nothing, simply ignore the signal
};

/* Close the driver if it has been already opened.
 */
void close_driver(void)
{
	if (drivernode != -1)
	{
		close(drivernode);
		drivernode = -1;
	};
};


/* Try to open the driver. If it is already open then we do nothing.
   RORCDRIVER_OK is returned if the open succeeds otherwise we return RORCDRIVER_CANT_OPEN_DRIVER
   and the details are returned by the system in "errno".
 */
rorcdriver_status_t open_driver(void)
{
	if (drivernode == -1)
	{
		drivernode = open(DRIVER_NODE, O_RDWR);  // open in read/write mode.
		if (drivernode == -1)
			return RORCDRIVER_CANT_OPEN_DRIVER;
		// Register the exit routine to clean up properly.
		if (registered_exit_routine == 0)
		{
			if ( atexit(&close_driver) != 0 )
			{
				close(drivernode);
				drivernode = -1;
				return RORCDRIVER_CANT_OPEN_DRIVER;
			};
		};
	};
	return RORCDRIVER_OK;
};

/* Open DMA buffers for reading and writing */

int * rorcdriverOpenEventBuffer(int length)
{
	rorcdriver_status_t status;
	status = open_driver();
	if (status != RORCDRIVER_OK)
		return NULL;
	else
	{
	        mmap_length = length;
		mmap_address = mmap(0, length, PROT_READ|PROT_WRITE, MAP_SHARED, drivernode, 0);
		return mmap_address;
	}
};

int rorcdriverGetReportWord(void)
{
	rorcdriver_status_t status;
	int report_word;

	status = open_driver();
	if (status == RORCDRIVER_OK)
	{
		ioctl(drivernode, rorcdriver_GETREPORTWORD, &report_word);
	}

        return report_word;
};

void rorcdriverProcessingCompleted(void)
{
	rorcdriver_status_t status;

	status = open_driver();
	if (status == RORCDRIVER_OK)
	{
		ioctl(drivernode, rorcdriver_PROCESSINGCOMPLETED);
	}
};

#ifdef DEBUG

rorcdriver_status_t rorcdriverReset(unsigned int irq)
{
	rorcdriver_status_t status;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, rorcdriver_UNREGISTER_ALL, irq);
};


rorcdriver_status_t rorcdriverResetAll(void)
{
	rorcdriver_status_t status;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, rorcdriver_UNREGISTER_ALL);
};


rorcdriver_status_t rorcdriverHardReset(void)
{
	rorcdriver_status_t status;
	status = open_driver();
	if (status != IRQSIG_OK)
		return status;
	else
		return ioctl(drivernode, rorcdriver_HARDRESET);
};




#endif // DEBUG


///////////////////////////////////////////////////////////////////////////////
// Implementation of API routines. Described in the header: "rorcdriver.h"

rorcdriver_status_t rorcdriverRegister(unsigned long event_buffer_length, unsigned long report_buffer_length, unsigned long irq_pulse_length, unsigned long event_length, unsigned long timeout_in_millisecs, int enable_dma_transfers, int enable_irq_transfers, int enable_pg_transfers)
{
	// We have to take care of the alarm signal scheduled in case of timeout
	struct sigaction request_timeout;
	request_timeout.sa_handler = request_timeout_handler;
	sigemptyset(&request_timeout.sa_mask);
	request_timeout.sa_flags = 0;
	int result = sigaction(SIGALRM, &request_timeout, 0);
	if (result != 0)
	{
		return 1;
	};


	rorcdriver_status_t status;
	rorcdriver_register_info_t info;

	info.event_buffer_length = (event_buffer_length << 2);
	info.report_buffer_length = (report_buffer_length << 2);
        info.irq_pulse_length = irq_pulse_length;
        info.event_length = event_length;
        info.timeout_in_millisecs = timeout_in_millisecs;
        info.enable_dma_transfers = enable_dma_transfers;
        info.enable_irq_transfers = enable_irq_transfers;
        info.enable_pg_transfers = enable_pg_transfers;

	status = open_driver();
	if (status != RORCDRIVER_OK)
		return status;
	else
		return ioctl(drivernode, rorcdriver_REGISTER, &info);
};


rorcdriver_status_t rorcdriverUnregister(void)
{
 munmap(mmap_address, mmap_length);

	rorcdriver_status_t status;
	rorcdriver_register_info_t info;
	status = open_driver();
	if (status != RORCDRIVER_OK)
		return status;
	else
		return ioctl(drivernode, rorcdriver_UNREGISTER, &info);
};


char* rorcdriverStatusString(rorcdriver_status_t status)
{
	switch (status)
	{
	case RORCDRIVER_OK:
		return "Successful";
	case RORCDRIVER_UNKNOWN_ERROR:
		return "Unknown error.";
	case RORCDRIVER_NO_SUCH_IOCTRL:
		return "Unrecognised ioctl number.";
	case RORCDRIVER_OUTOFMEM:
		return "System ran out of memory.";
	case RORCDRIVER_CANT_SCHEDULE_TASK:
		return "Can not schedule the dispatcher task.";
	case RORCDRIVER_CANT_OPEN_DRIVER:
		return "Can not open driver file.";
	default:
		return strerror(status);
	};
};
